import difflib
import types
from collections.abc import Callable, Iterator
from functools import cached_property, partial, singledispatchmethod
from itertools import islice
from pathlib import Path
from typing import TYPE_CHECKING, Any

import click
from eth_pydantic_types import HexBytes
from eth_utils import to_hex
from ethpm_types.abi import EventABI

from ape.api.accounts import AccountAPI
from ape.api.address import Address, BaseAddress
from ape.api.query import (
    ContractCreation,
    ContractEventQuery,
    extract_fields,
    validate_and_expand_columns,
)
from ape.exceptions import (
    ApeAttributeError,
    ArgumentsLengthError,
    ChainError,
    ContractDataError,
    ContractLogicError,
    ContractNotFoundError,
    CustomError,
    MethodNonPayableError,
    MissingDeploymentBytecodeError,
)
from ape.logging import get_rich_console, logger
from ape.types.events import ContractLog, LogFilter, MockContractLog
from ape.utils.abi import StructParser, _enrich_natspec, encode_topics, is_array
from ape.utils.basemodel import (
    BaseInterfaceModel,
    ExtraAttributesMixin,
    ExtraModelAttributes,
    ManagerAccessMixin,
    _assert_not_ipython_check,
    get_attribute_with_extras,
    only_raise_attribute_error,
)
from ape.utils.misc import log_instead_of_fail

if TYPE_CHECKING:
    from ethpm_types.abi import ConstructorABI, ErrorABI, MethodABI
    from ethpm_types.contract_type import ABI_W_SELECTOR_T, ContractType
    from pandas import DataFrame

    from ape.api.networks import ProxyInfoAPI
    from ape.api.providers import CallResult
    from ape.api.transactions import ReceiptAPI, TransactionAPI
    from ape.types.address import AddressType


class ContractConstructor(ManagerAccessMixin):
    def __init__(
        self,
        deployment_bytecode: HexBytes,
        abi: "ConstructorABI",
    ) -> None:
        self.deployment_bytecode = deployment_bytecode
        self.abi = abi

        if not self.deployment_bytecode:
            logger.warning("Deploying an empty contract (no bytecode)")
            self.deployment_bytecode = HexBytes("")

    @log_instead_of_fail(default="<ContractConstructor>")
    def __repr__(self) -> str:
        return self.abi.signature if self.abi else "constructor()"

    def encode_input(self, *args) -> HexBytes:
        ecosystem = self.provider.network.ecosystem
        encoded_calldata = ecosystem.encode_calldata(self.abi, *args)
        return HexBytes(encoded_calldata)

    def decode_input(self, calldata: bytes) -> tuple[str, dict[str, Any]]:
        decoded_inputs = self.provider.network.ecosystem.decode_calldata(self.abi, calldata)
        return self.abi.selector, decoded_inputs

    def serialize_transaction(self, *args, **kwargs) -> "TransactionAPI":
        arguments = self.conversion_manager.convert_method_args(self.abi, args)
        converted_kwargs = self.conversion_manager.convert_method_kwargs(kwargs)
        return self.provider.network.ecosystem.encode_deployment(
            self.deployment_bytecode, self.abi, *arguments, **converted_kwargs
        )

    def __call__(self, private: bool = False, *args, **kwargs) -> "ReceiptAPI":
        txn = self.serialize_transaction(*args, **kwargs)

        if "sender" in kwargs and hasattr(kwargs["sender"], "call"):
            sender = kwargs["sender"]
            return sender.call(txn, **kwargs)
        elif "sender" not in kwargs and self.account_manager.default_sender is not None:
            return self.account_manager.default_sender.call(txn, **kwargs)

        return (
            self.provider.send_private_transaction(txn)
            if private
            else self.provider.send_transaction(txn)
        )


class ContractCall(ManagerAccessMixin):
    def __init__(self, abi: "MethodABI", address: "AddressType") -> None:
        super().__init__()
        self.abi = abi
        self.address = address

    @log_instead_of_fail(default="<ContractCall>")
    def __repr__(self) -> str:
        return self.abi.signature

    def serialize_transaction(self, *args, **kwargs) -> "TransactionAPI":
        converted_kwargs = self.conversion_manager.convert_method_kwargs(kwargs)
        return self.provider.network.ecosystem.encode_transaction(
            self.address, self.abi, *args, **converted_kwargs
        )

    def __call__(self, *args, **kwargs) -> Any:
        decode_output = kwargs.pop("decode", True)
        txn = self.serialize_transaction(*args, **kwargs)
        txn.chain_id = self.provider.network.chain_id

        result = self.provider.send_call(txn, **kwargs)

        # If the result is raw bytes, handle it directly.
        if isinstance(result, HexBytes):
            return self._decode_returndata(result) if decode_output else result

        # At this point, result must be a CallResult.
        # Note: type hinting here is for clarity only.
        call_result: CallResult = result  # type: ignore

        if not decode_output:
            return call_result.returndata

        elif call_result.revert:
            return call_result.revert.revert_message

        return call_result.decode(self.abi)

    def _decode_returndata(self, returndata: HexBytes) -> Any:
        # The current and main way of decoding call-results (HexBytes) before
        # the wider adoption of the `CallResult` class.
        # Decode the output bytes into Python types.
        decoded_output = self.provider.network.ecosystem.decode_returndata(
            self.abi,
            returndata,
        )

        if not isinstance(decoded_output, (list, tuple)):
            return decoded_output

        # NOTE: Returns a tuple, so make sure to handle all the cases
        elif len(decoded_output) < 2:
            return decoded_output[0] if len(decoded_output) == 1 else None

        return decoded_output


class ContractMethodHandler(ManagerAccessMixin):
    contract: "ContractInstance"
    abis: list["MethodABI"]

    def __init__(self, contract: "ContractInstance", abis: list["MethodABI"]) -> None:
        super().__init__()
        self.contract = contract
        self.abis = abis

        # If there is a natspec, inject it as the "doc-str" for this method.
        # This greatly helps integrate with IPython.
        self.__doc__ = self.info

    @log_instead_of_fail(default="<ContractMethodHandler>")
    def __repr__(self) -> str:
        # `<ContractName 0x1234...AbCd>.method_name`
        return f"{self.contract.__repr__()}.{self.abis[-1].name}"

    @log_instead_of_fail()
    def _repr_pretty_(self, printer, cycle):
        """
        Show the NatSpec of a Method in any IPython console (including ``ape console``).
        """
        console = get_rich_console()
        output = self._get_info(enrich=True) or "\n".join(abi.signature for abi in self.abis)
        console.print(output)

    def __str__(self) -> str:
        # `method_name(type1 arg1, ...) -> return_type`
        abis = sorted(self.abis, key=lambda abi: len(abi.inputs or []))
        return abis[-1].signature

    @property
    def info(self) -> str:
        """
        The NatSpec documentation of the method, if one exists.
        Else, returns the empty string.
        """
        return self._get_info()

    def _get_info(self, enrich: bool = False) -> str:
        infos: list[str] = []

        for abi in self.abis:
            if abi.selector not in self.contract.contract_type.natspecs:
                continue

            natspec = self.contract.contract_type.natspecs[abi.selector]
            header = abi.signature
            natspec_str = natspec.replace("\n", "\n  ")
            infos.append(f"{header}\n  {natspec_str}")

        if enrich:
            infos = [_enrich_natspec(n) for n in infos]

        # Same as number of ABIs, regardless of NatSpecs.
        number_infos = len(infos)
        if number_infos == 1:
            return infos[0]

        # Ensure some distinction of the infos using number-prefixes.
        numeric_infos = []
        for idx, info in enumerate(infos):
            num_info = f"{idx + 1}: {info}"
            numeric_infos.append(num_info)

        return "\n\n".join(numeric_infos)

    def encode_input(self, *args) -> HexBytes:
        selected_abi = _select_method_abi(self.abis, args)
        arguments = self.conversion_manager.convert_method_args(selected_abi, args)
        ecosystem = self.provider.network.ecosystem
        encoded_calldata = ecosystem.encode_calldata(selected_abi, *arguments)
        method_id = ecosystem.get_method_selector(selected_abi)
        return HexBytes(method_id + encoded_calldata)

    def decode_input(self, calldata: bytes) -> tuple[str, dict[str, Any]]:
        matching_abis = []
        rest_calldata = None
        err = ContractDataError(
            f"Unable to find matching method ABI for calldata '{to_hex(calldata)}'. "
            "Try prepending a method ID to the beginning of the calldata."
        )
        for abi in self.abis:
            selector = self.provider.network.ecosystem.get_method_selector(abi)
            if calldata.startswith(selector):
                cutoff = len(selector)
                rest_calldata = calldata[cutoff:]
                matching_abis.append(abi)

        if len(matching_abis) == 1:
            abi = matching_abis[0]
            decoded_input = self.provider.network.ecosystem.decode_calldata(
                matching_abis[0], HexBytes(rest_calldata or "")
            )
            return abi.selector, decoded_input

        elif len(matching_abis) > 1:
            raise err

        # Brute-force find method ABI
        valid_results = []
        for abi in self.abis:
            try:
                decoded_calldata = self.provider.network.ecosystem.decode_calldata(
                    abi, HexBytes(calldata)
                )
            except Exception:
                continue

            if decoded_calldata:
                valid_results.append((abi, decoded_calldata))

        if len(valid_results) == 1:
            selected_abi, decoded_calldata = valid_results[0]
            return selected_abi.selector, decoded_calldata

        raise err

    def _validate_is_contract(self):
        if not self.contract.is_contract:
            raise ContractNotFoundError(
                self.contract.address,
                self.provider.network.explorer is not None,
                self.provider.network_choice,
            )


class ContractCallHandler(ContractMethodHandler):
    def __call__(self, *args, **kwargs) -> Any:
        self._validate_is_contract()
        selected_abi = _select_method_abi(self.abis, args)
        arguments = self.conversion_manager.convert_method_args(selected_abi, args)

        return ContractCall(
            abi=selected_abi,
            address=self.contract.address,
        )(*arguments, **kwargs)

    def as_transaction(self, *args, **kwargs):
        """
        Convert the call to a transaction. This is useful for checking coverage
        or checking gas costs.

        Args:
            *args: The contract method invocation arguments.
            **kwargs: Transaction kwargs, such as value or
              sender.

        Returns:
            :class:`~ape.api.transactions.TransactionAPI`
        """
        return self.transact.as_transaction(*args, **kwargs)

    def as_transaction_bytes(self, *args, **txn_kwargs) -> HexBytes:
        """
        Get a signed serialized transaction.

        Returns:
            HexBytes: The serialized transaction
        """
        return self.transact.as_transaction_bytes(**txn_kwargs)

    @property
    def transact(self) -> "ContractTransactionHandler":
        """
        Send the call as a transaction.
        """
        return ContractTransactionHandler(self.contract, self.abis)

    def estimate_gas_cost(self, *args, **kwargs) -> int:
        """
        Get the estimated gas cost (according to the provider) for the
        contract method call (as if it were a transaction).

        Args:
            *args: The contract method invocation arguments.
            **kwargs: Transaction kwargs, such as value or
              sender.

        Returns:
            int: The estimated cost of gas to execute the transaction
            reported in the fee-currency's smallest unit, e.g. Wei.
        """

        selected_abi = _select_method_abi(self.abis, args)
        arguments = self.conversion_manager.convert_method_args(selected_abi, args)
        return self.transact.estimate_gas_cost(*arguments, **kwargs)


def _select_method_abi(abis: list["MethodABI"], args: tuple | list) -> "MethodABI":
    args = args or []
    selected_abi = None
    for abi in abis:
        inputs = abi.inputs or []
        if len(args) == len(inputs):
            selected_abi = abi

    if not selected_abi:
        raise ArgumentsLengthError(len(args), inputs=abis)

    return selected_abi


class ContractTransaction(ManagerAccessMixin):
    def __init__(self, abi: "MethodABI", address: "AddressType") -> None:
        super().__init__()
        self.abi: MethodABI = abi
        self.address: AddressType = address

    @log_instead_of_fail(default="<ContractTransaction>")
    def __repr__(self) -> str:
        return self.abi.signature

    def serialize_transaction(self, *args, **kwargs) -> "TransactionAPI":
        if "sender" in kwargs and isinstance(kwargs["sender"], (ContractInstance, Address)):
            # Automatically impersonate contracts (if API available) when sender
            kwargs["sender"] = self.account_manager.test_accounts[kwargs["sender"].address]

        arguments = self.conversion_manager.convert_method_args(self.abi, args)
        converted_kwargs = self.conversion_manager.convert_method_kwargs(kwargs)
        return self.provider.network.ecosystem.encode_transaction(
            self.address, self.abi, *arguments, **converted_kwargs
        )

    def __call__(self, *args, **kwargs) -> "ReceiptAPI":
        txn = self.serialize_transaction(*args, **kwargs)
        private = kwargs.get("private", False)

        if "sender" in kwargs and hasattr(kwargs["sender"], "call"):
            return kwargs["sender"].call(txn, **kwargs)

        txn = self.provider.prepare_transaction(txn)
        return (
            self.provider.send_private_transaction(txn)
            if private
            else self.provider.send_transaction(txn)
        )


class ContractTransactionHandler(ContractMethodHandler):
    def as_transaction(self, *args, **kwargs) -> "TransactionAPI":
        """
        Get a :class:`~ape.api.transactions.TransactionAPI`
        for this contract method invocation. This is useful
        for simulations or estimating fees without sending
        the transaction.

        Args:
            *args: The contract method invocation arguments.
            **kwargs: Transaction kwargs, such as value or
              sender.

        Returns:
            :class:`~ape.api.transactions.TransactionAPI`
        """
        sign = kwargs.pop("sign", False)
        contract_transaction = self._as_transaction(*args)
        transaction = contract_transaction.serialize_transaction(*args, **kwargs)
        self.provider.prepare_transaction(transaction)

        if sender := kwargs.get("sender"):
            if isinstance(sender, AccountAPI):
                prepped_tx = sender.prepare_transaction(transaction)
                return (sender.sign_transaction(prepped_tx) or prepped_tx) if sign else prepped_tx

        return transaction

    def as_transaction_bytes(self, *args, **txn_kwargs) -> HexBytes:
        """
        Get a signed serialized transaction.

        Returns:
            HexBytes: The serialized transaction
        """
        txn_kwargs["sign"] = True
        tx = self.as_transaction(*args, **txn_kwargs)
        return tx.serialize_transaction()

    def estimate_gas_cost(self, *args, **kwargs) -> int:
        """
        Get the estimated gas cost (according to the provider) for the
        contract method-invocation transaction.

        Args:
            *args: The contract method invocation arguments.
            **kwargs: Transaction kwargs, such as value or
              sender.

        Returns:
            int: The estimated cost of gas to execute the transaction
            reported in the fee-currency's smallest unit, e.g. Wei.
        """
        selected_abi = _select_method_abi(self.abis, args)
        arguments = self.conversion_manager.convert_method_args(selected_abi, args)
        txn = self.as_transaction(*arguments, **kwargs)
        return self.provider.estimate_gas_cost(txn)

    @property
    def call(self) -> ContractCallHandler:
        """
        Get the :class:`~ape.contracts.base.ContractCallHandler` equivalent
        of this transaction handler. The call-handler uses the ``eth_call``
        RPC under-the-hood and thus it gets reverted before submitted.
        This is a useful way to simulate a transaction without invoking it.
        """

        return ContractCallHandler(self.contract, self.abis)

    def __call__(self, *args, **kwargs) -> "ReceiptAPI":
        contract_transaction = self._as_transaction(*args)
        if "sender" not in kwargs and self.account_manager.default_sender is not None:
            kwargs["sender"] = self.account_manager.default_sender

        return contract_transaction(*args, **kwargs)

    def _as_transaction(self, *args) -> ContractTransaction:
        self._validate_is_contract()
        selected_abi = _select_method_abi(self.abis, args)
        return ContractTransaction(
            abi=selected_abi,
            address=self.contract.address,
        )


# TODO: In Ape 0.9 - make not a BaseModel - no reason to.
class ContractEvent(BaseInterfaceModel):
    """
    The types of events on a :class:`~ape.contracts.base.ContractInstance`.
    Use the event types via ``.`` access on the contract instances.

    Usage example::

         # 'my_contract' refers to a ContractInstance in this case.
         my_event_type = my_contract.MyEvent
    """

    contract: "ContractTypeWrapper"
    abi: EventABI
    _logs: list[ContractLog] | None = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Inject the doc-str using the NatSpec to better integrate with IPython.
        # NOTE: This must happen AFTER super().__init__().
        self.__doc__ = self.info

    @log_instead_of_fail(default="<ContractEvent>")
    def __repr__(self) -> str:
        return self.abi.signature

    @log_instead_of_fail()
    def _repr_pretty_(self, printer, cycle):
        console = get_rich_console()
        console.print(self._get_info(enrich=True))

    @property
    def name(self) -> str:
        """
        The name of the contract event, as defined in the contract.
        """
        return self.abi.name

    @property
    def info(self) -> str:
        """
        NatSpec info derived from the contract-type developer-documentation.
        """
        return self._get_info()

    def _get_info(self, enrich: bool = False) -> str:
        info_str = self.abi.signature
        if spec := self.contract.contract_type.natspecs.get(self.abi.selector):
            spec_indented = spec.replace("\n", "\n  ")
            info_str = f"{info_str}\n  {spec_indented}"

        return _enrich_natspec(info_str) if enrich else info_str

    def __iter__(self) -> Iterator[ContractLog]:  # type: ignore[override]
        """
        Get all logs that have occurred for this event.
        """
        yield from self.range(self.chain_manager.blocks.height + 1)

    @property
    def log_filter(self) -> LogFilter:
        # NOTE: This shouldn't really be called when given contract containers.
        address = getattr(self.contract, "address", None)
        addresses = [] if not address else [address]
        return LogFilter.from_event(event=self.abi, addresses=addresses, start_block=0)

    @singledispatchmethod
    def __getitem__(self, value) -> ContractLog | list[ContractLog]:  # type: ignore[override]
        raise NotImplementedError(f"Cannot use '{type(value)}' to access logs.")

    @__getitem__.register
    def __getitem_int(self, index: int) -> ContractLog:
        """
        Access events on the contract by the index of when they occurred.

        Args:
            index (int): The index such that ``0`` is the first log to have occurred
              and ``-1`` is the last.

        Returns:
            :class:`~ape.contracts.base.ContractLog`
        """
        logs = self.provider.get_contract_logs(self.log_filter)
        try:
            if index == 0:
                return next(logs)
            elif index > 0:
                return next(islice(logs, index, index + 1))
            else:
                return list(logs)[index]

        except (IndexError, StopIteration) as err:
            raise IndexError(f"No log at index '{index}' for event '{self.abi.name}'.") from err

    @__getitem__.register
    def __getitem_slice(self, value: slice) -> list[ContractLog]:
        """
        Access a slice of logs from this event.

        Args:
            value (slice): The range of log to get, e.g. ``[5:10]``.

        Returns:
            Iterator[:class:`~ape.contracts.base.ContractLog`]
        """
        logs = self.provider.get_contract_logs(self.log_filter)
        return list(islice(logs, value.start, value.stop, value.step))

    def __len__(self):
        logs = self.provider.get_contract_logs(self.log_filter)
        return sum(1 for _ in logs)

    def __call__(self, *args: Any, **kwargs: Any) -> MockContractLog:
        """
        Create a mock-instance of a log using this event ABI and the contract address.

        Args:
            *args: Positional arguments for the event.
            **kwargs: Key-word arguments for the event.

        Returns:
            :class:`~ape.types.events.MockContractLog`
        """
        # Create a dictionary from the positional arguments.
        #
        # NOTE: We intentionally allow providing only a subset of positional args
        # (tests and user code often construct partial mock logs).
        #
        # However, we still want to fail fast if *too many* positional args are
        # provided. Using zip(strict=True) with an inputs iterator limited to
        # len(args) keeps the "too many args" safety without breaking the
        # "partial args" use-case.
        input_names = (ipt.name for ipt in islice(self.abi.inputs, len(args)))
        event_args: dict[Any, Any] = dict(zip(input_names, args, strict=True))
        if overlapping_keys := set(event_args).intersection(kwargs):
            raise ValueError(
                f"Overlapping keys found in arguments: '{', '.join(overlapping_keys)}'."
            )

        # Update event_args with keyword arguments
        event_args.update(kwargs)

        # Check that event_args.keys() is a subset of the expected input names
        keys_given = set(event_args.keys())
        keys_expected = {ipt.name for ipt in self.abi.inputs}
        if unknown_input_names := keys_given - keys_expected:
            message = "Unknown keys: "
            sections = []
            for unknown in unknown_input_names:
                if matches := difflib.get_close_matches(unknown, keys_expected, n=1, cutoff=0.5):
                    matches_str = ", ".join(matches)
                    sections.append(f"{unknown} (did you mean: '{matches_str}'?)")
                else:
                    sections.append(unknown)

            message = f"{message} '{', '.join(sections)}'"
            raise ValueError(message)

        # Convert the arguments using the conversion manager
        converted_args = {}
        ecosystem = self.provider.network.ecosystem
        parser = StructParser(self.abi)

        for key, value in event_args.items():
            if value is None:
                continue

            input_abi = next(ipt for ipt in self.abi.inputs if ipt.name == key)
            py_type = ecosystem.get_python_types(input_abi)
            if isinstance(value, dict):
                if "tuple" in str(input_abi.type) and input_abi.components:
                    # Preserve struct component order and ignore extra keys.
                    ls_values = [value.get(c.name) for c in input_abi.components if c.name]
                    converted_values = self.conversion_manager.convert(ls_values, py_type)

                    # NOTE: eth-abi returns tuple-addresses as lower-case strings.
                    # Match that behavior so mock logs compare equal with decoded logs.
                    for i, comp in enumerate(input_abi.components):
                        if (
                            comp.canonical_type == "address"
                            and i < len(converted_values)
                            and isinstance(converted_values[i], str)
                        ):
                            converted_values[i] = converted_values[i].lower()

                    converted_args[key] = parser.decode_input([converted_values])

                else:
                    ls_values = list(value.values())
                    converted_values = self.conversion_manager.convert(ls_values, py_type)
                    converted_args[key] = parser.decode_input([converted_values])

            elif isinstance(value, (list, tuple)):
                # Arrays are represented as lists/tuples in Python.
                # Don't pass them through StructParser (which is for tuples/structs).
                converted_args[key] = (
                    self.conversion_manager.convert(value, py_type)
                    if is_array(input_abi.type)
                    else parser.decode_input(value)
                )

            else:
                converted_args[key] = self.conversion_manager.convert(value, py_type)

        properties: dict = {
            "event_arguments": converted_args,
            "event_name": self.abi.name,
        }
        if hasattr(self.contract, "address"):
            # Only address if this is off an instance.
            properties["contract_address"] = self.contract.address

        return MockContractLog(**properties)

    def query(
        self,
        *columns: str,
        start_block: int = 0,
        stop_block: int | None = None,
        step: int = 1,
        engine_to_use: str | None = None,
    ) -> "DataFrame":
        """
        Iterate through blocks for log events

        Args:
            *columns (str): ``*``-based argument for columns in the DataFrame to
              return.
            start_block (int): The first block, by number, to include in the
              query. Defaults to ``0``.
            stop_block (int | None): The last block, by number, to include
              in the query. Defaults to the latest block.
            step (int): The number of blocks to iterate between block numbers.
              Defaults to ``1``.
            engine_to_use (str | None): query engine to use, bypasses query
              engine selection algorithm.

        Returns:
            pd.DataFrame
        """
        # perf: pandas import is really slow. Avoid importing at module level.
        import pandas as pd

        HEAD = self.chain_manager.blocks.height
        if start_block < 0:
            start_block = HEAD + start_block

        if stop_block is None:
            stop_block = HEAD

        elif stop_block < 0:
            stop_block = HEAD + stop_block

        elif stop_block > HEAD:
            raise ChainError(
                f"'stop={stop_block}' cannot be greater than the chain length ({HEAD})."
            )
        query: dict = {
            "columns": (list(ContractLog.__pydantic_fields__) if columns[0] == "*" else columns),
            "event": self.abi,
            "start_block": start_block,
            "stop_block": stop_block,
            "step": step,
        }
        if hasattr(self.contract, "address"):
            # Only query for a specific contract when checking an instance.
            query["contract"] = self.contract.address

        contract_event_query = ContractEventQuery(**query)
        contract_events = self.query_manager.query(
            contract_event_query, engine_to_use=engine_to_use
        )
        columns_ls = validate_and_expand_columns(columns, ContractLog)
        data = map(partial(extract_fields, columns=columns_ls), contract_events)
        return pd.DataFrame(columns=columns_ls, data=data)

    def range(
        self,
        start_or_stop: int,
        stop: int | None = None,
        search_topics: dict[str, Any] | None = None,
        extra_addresses: list | None = None,
    ) -> Iterator[ContractLog]:
        """
        Search through the logs for this event using the given filter parameters.

        Args:
            start_or_stop (int): When also given ``stop``, this is the earliest
              block number in the desired log set.
              Otherwise, it is the total amount of blocks to get starting from ``0``.
            stop (int | None): The latest block number in the
              desired log set. Defaults to delegating to provider.
            search_topics (dict | None): Search topics, such as indexed event inputs,
              to query by. Defaults to getting all events.
            extra_addresses (list | None): Additional contract addresses containing the same event type.
              Defaults to only looking at the contract instance where this event is defined.

        Returns:
            Iterator[:class:`~ape.contracts.base.ContractLog`]
        """
        if not (contract_address := getattr(self.contract, "address", None)):
            return

        start_block = None
        stop_block = None
        HEAD = self.chain_manager.blocks.height  # Current block height

        if stop is None:
            contract = None
            try:
                contract = self.chain_manager.contracts.instance_at(contract_address)
            except Exception:
                pass

            # Determine the start block from contract creation metadata
            if contract and (creation := contract.creation_metadata):
                start_block = creation.block

            # Handle single parameter usage (like Python's range(stop))
            if start_or_stop == 0:
                # stop==0 is the same as stop==HEAD
                # because of the -1 (turns to negative).
                stop_block = HEAD + 1
            elif start_or_stop >= 0:
                # Given like range(1)
                stop_block = min(start_or_stop - 1, HEAD)
            else:
                # Give like range(-1)
                stop_block = HEAD + start_or_stop

        elif start_or_stop is not None and stop is not None:
            # Handle cases where both start and stop are provided
            if start_or_stop >= 0:
                start_block = min(start_or_stop, HEAD)
            else:
                # Negative start relative to HEAD
                adjusted_value = HEAD + start_or_stop + 1
                start_block = max(adjusted_value, 0)

            if stop == 0:
                # stop==0 is the same as stop==HEAD
                # because of the -1 (turns to negative).
                stop_block = HEAD
            elif stop > 0:
                # Positive stop, capped to the chain HEAD
                stop_block = min(stop - 1, HEAD)
            else:
                # Negative stop.
                adjusted_value = HEAD + stop
                stop_block = max(adjusted_value, 0)

        # Gather all addresses to query (contract and any extra ones provided)
        addresses = list(set([contract_address] + (extra_addresses or [])))

        # Construct the event query
        contract_event_query = ContractEventQuery(
            columns=list(ContractLog.__pydantic_fields__),  # Ensure all necessary columns
            contract=addresses,
            event=self.abi,
            search_topics=search_topics,
            start_block=start_block or 0,  # Default to block 0 if not set
            stop_block=stop_block,  # None means query to the current HEAD
        )

        # Execute the query and yield results
        yield from self.query_manager.query(contract_event_query)  # type: ignore

    def from_receipt(self, receipt: "ReceiptAPI") -> list[ContractLog]:
        """
        Get all the events from the given receipt.

        Args:
            receipt (:class:`~ape.api.transactions.ReceiptAPI`): The receipt containing the logs.

        Returns:
            list[:class:`~ape.contracts.base.ContractLog`]
        """
        ecosystem = self.provider.network.ecosystem

        # NOTE: Safe to use a list because a receipt should never have too many logs.
        return list(
            ecosystem.decode_logs(
                # NOTE: Filter out logs that do not match this address (if address available)
                #       (okay to be empty list, since EcosystemAPI.decode_logs will return [])
                [
                    log
                    for log in receipt.logs
                    if log["address"] == getattr(self.contract, "address", log["address"])
                ],
                self.abi,
            )
        )

    def poll_logs(
        self,
        start_block: int | None = None,
        stop_block: int | None = None,
        required_confirmations: int | None = None,
        new_block_timeout: int | None = None,
        search_topics: dict[str, Any] | None = None,
        **search_topic_kwargs: dict[str, Any],
    ) -> Iterator[ContractLog]:
        """
        Poll new logs for this event. Can specify topic filters to use when setting up the filter.
        Optionally set a start block to include historical blocks.

        **NOTE**: This is a daemon method; it does not terminate unless an exception occurs.

        Usage example::

            for new_log in contract.MyEvent.poll_logs(arg=1):
                print(f"New event log found: block_number={new_log.block_number}")

        Args:
            start_block (int | None): The block number to start with. Defaults to the pending
              block number.
            stop_block (int | None): Optionally set a future block number to stop at.
              Defaults to never-ending.
            required_confirmations (int | None): The amount of confirmations to wait
              before yielding the block. The more confirmations, the less likely a reorg will occur.
              Defaults to the network's configured required confirmations.
            new_block_timeout (int | None): The amount of time to wait for a new block before
              quitting. Defaults to 10 seconds for local networks or ``50 * block_time`` for live
              networks.
            search_topics (dict[str, Any] | None): A dictionary of search topics to use when
              constructing a polling filter. Overrides the value of `**search_topics_kwargs`.
            search_topics_kwargs: Search topics to use when constructing a polling filter. Allows
              easily specifying topic filters using kwarg syntax but can be used w/ `search_topics`

        Returns:
            Iterator[:class:`~ape.types.ContractLog`]
        """

        if required_confirmations is None:
            required_confirmations = self.provider.network.required_confirmations

        # NOTE: We process historical blocks separately here to minimize rpc calls
        height = max(self.chain_manager.blocks.height - required_confirmations, 0)
        if start_block and height > 0 and start_block < height:
            yield from self.range(start_block, height)
            start_block = height + 1

        if search_topics:
            search_topic_kwargs.update(search_topics)

        topics = encode_topics(self.abi, search_topic_kwargs)

        if address := getattr(self.contract, "address", None):
            # NOTE: Now we process the rest
            yield from self.provider.poll_logs(
                stop_block=stop_block,
                address=address,
                required_confirmations=required_confirmations,
                new_block_timeout=new_block_timeout,
                events=[self.abi],
                topics=topics,
            )


# TODO: In 0.9, just make `_events_` or ContractEvent possibly handle multiple ABIs
#   much like the transactions handlers do. OR at least take the opportunity to refactor.
class ContractEventWrapper:
    """
    A wrapper used when multiple events have the same so that
    you can still create mock-logs.
    """

    def __init__(self, events: list[ContractEvent]):
        self.events = events

    def __call__(self, *args, **kwargs) -> MockContractLog:
        """
        Create a mock contract log using the first working ABI.

        Args:
            *args: Positional arguments for the event.
            **kwargs: Key-word arguments for the event.

        Returns:
            :class:`~ape.types.events.MockContractLog`
        """
        # TODO: Use composite error.
        errors = []
        for evt in self.events:
            try:
                return evt(*args, **kwargs)
            except ValueError as err:
                errors.append(err)
                continue  # not a match

        error_str = ", ".join([f"{e}" for e in errors])
        raise ValueError(f"Could not make a mock contract log. Errors: {error_str}")


class ContractTypeWrapper(ManagerAccessMixin):
    contract_type: "ContractType"
    base_path: Path | None = None

    @property
    def selector_identifiers(self) -> dict[str, str]:
        """
        Provides a mapping of function signatures (pre-hashed selectors) to
        selector identifiers.
        """
        return self.contract_type.selector_identifiers

    @property
    def identifier_lookup(self) -> dict[str, "ABI_W_SELECTOR_T"]:
        """
        Provides a mapping of method, error, and event selector identifiers to
        ABI Types.
        """
        return self.contract_type.identifier_lookup

    @property
    def source_path(self) -> Path | None:
        """
        Returns the path to the local contract if determined that this container
        belongs to the active project by cross-checking source_id.
        """
        if not (source_id := self.contract_type.source_id):
            return None

        base = self.base_path or self.local_project.path
        path = base / source_id
        return path if path.is_file() else None

    def decode_input(self, calldata: bytes) -> tuple[str, dict[str, Any]]:
        """
        Decode the given calldata using this contract.
        If the calldata has a method ID prefix, Ape will detect it and find
        the corresponding method, else it will error.

        Args:
            calldata (bytes): The calldata to decode.

        Returns:
            tuple[str, dict[str, Any]]: A tuple containing the method selector
            along a mapping of input names to their decoded values.
            If an input does not have a number, it will have the stringified
            index as its key.
        """

        ecosystem = self.provider.network.ecosystem
        if calldata in self.contract_type.mutable_methods:
            method = self.contract_type.mutable_methods[calldata]
        elif calldata in self.contract_type.view_methods:
            method = self.contract_type.view_methods[calldata]
        else:
            method = None

        if not method:
            raise ContractDataError(
                f"Unable to find method ABI from calldata '{to_hex(calldata)}'. "
                "Try prepending the method ID to the beginning of the calldata."
            )

        method_id = ecosystem.get_method_selector(method)
        cutoff = len(method_id)
        rest_calldata = calldata[cutoff:]
        input_dict = ecosystem.decode_calldata(method, rest_calldata)
        return method.selector, input_dict

    def _create_custom_error_type(self, abi: "ErrorABI", **kwargs) -> type[CustomError]:
        def exec_body(namespace):
            namespace["abi"] = abi
            namespace["contract"] = self
            for key, val in kwargs.items():
                namespace[key] = val

        error_type = types.new_class(abi.name, (CustomError,), {}, exec_body)
        natspecs = self.contract_type.natspecs

        def _get_info(enrich: bool = False) -> str:
            if not (natspec := natspecs.get(abi.selector)):
                return ""

            elif enrich:
                natspec = _enrich_natspec(natspec)

            return f"{abi.signature}\n  {natspec}"

        def _repr_pretty_(cls, *args, **kwargs):
            console = get_rich_console()
            output = _get_info(enrich=True) or repr(cls)
            console.print(output)

        def repr_pretty_for_assignment(cls, *args, **kwargs):
            return _repr_pretty_(error_type, *args, **kwargs)

        info = _get_info()
        error_type.info = error_type.__doc__ = info  # type: ignore
        if info:
            # perf: Avoid forcing everyone to import from IPython.
            from IPython.lib.pretty import for_type

            error_type._repr_pretty_ = repr_pretty_for_assignment  # type: ignore

            # Register the dynamically-created type with IPython so it integrates.
            for_type(type(error_type), _repr_pretty_)

        return error_type


class ContractInstance(BaseAddress, ContractTypeWrapper):
    """
    An interactive instance of a smart contract.
    After you deploy a contract using the :class:`~ape.api.accounts.AccountAPI.deploy` method,
    you get back a contract instance.

    Usage example::

        from ape import accounts, project

        a = accounts.load("alias")  # Load an account by alias
        contract = a.deploy(project.MyContract)  # The result of 'deploy()' is a ContractInstance
    """

    def __init__(
        self,
        address: "AddressType",
        contract_type: "ContractType",
        txn_hash: str | HexBytes | None = None,
    ) -> None:
        super().__init__()
        self._address = address
        self.contract_type = contract_type
        self.txn_hash = (
            (txn_hash if isinstance(txn_hash, str) else to_hex(txn_hash)) if txn_hash else None
        )

    def __call__(self, *args, **kwargs) -> "ReceiptAPI":
        has_value = kwargs.get("value")
        has_data = kwargs.get("data") or kwargs.get("input")
        has_non_payable_fallback = (
            self.contract_type.fallback and not self.contract_type.fallback.is_payable
        )

        if has_value and has_non_payable_fallback and self.contract_type.receive is None:
            # User is sending a value when the contract doesn't accept it.
            raise MethodNonPayableError(
                "Contract's fallback is non-payable and there is no receive ABI. "
                "Unable to send value."
            )

        elif has_value and has_data and has_non_payable_fallback:
            # User is sending both value and data. When sending data, the fallback
            # is always triggered. Thus, since it is non-payable, it would fail.
            raise MethodNonPayableError(
                "Sending both value= and data= but fallback is non-payable."
            )

        return super().__call__(*args, **kwargs)

    @classmethod
    def from_receipt(
        cls, receipt: "ReceiptAPI", contract_type: "ContractType"
    ) -> "ContractInstance":
        """
        Create a contract instance from the contract deployment receipt.
        """
        address = receipt.contract_address
        if not address:
            raise ChainError(
                "Receipt missing 'contract_address' field. "
                "Was this from a deploy transaction (e.g. `project.MyContract.deploy()`)?"
            )

        instance = cls(
            address=address,
            contract_type=contract_type,
            txn_hash=receipt.txn_hash,
        )

        # Cache creation.
        creation = ContractCreation.from_receipt(receipt)
        cls.chain_manager.contracts.contract_creations[address] = creation

        return instance

    @property
    def creation_metadata(self) -> ContractCreation | None:
        """
        Contract creation details: txn_hash, block, deployer, factory, receipt.
        See :class:`~ape.api.query.ContractCreation` for more details.
        **NOTE**: Must be either connected to a node that provides this data,
        such as a node with the ``ots_`` namespace enabled, or have a query-plugin
        installed that can fetch this data, such as ``ape-etherscan``.
        """
        return self.chain_manager.contracts.get_creation_metadata(self.address)

    @log_instead_of_fail(default="<ContractInstance>")
    def __repr__(self) -> str:
        contract_name = self.contract_type.name or "Unnamed contract"
        return f"<{contract_name} {self.address}>"

    @property
    def address(self) -> "AddressType":
        """
        The address of the contract.

        Returns:
            :class:`~ape.types.address.AddressType`
        """
        return self._address

    @cached_property
    def _view_methods_(self) -> dict[str, ContractCallHandler]:
        view_methods: dict[str, list[MethodABI]] = dict()

        for abi in self.contract_type.view_methods:
            if abi.name in view_methods:
                view_methods[abi.name].append(abi)
            else:
                view_methods[abi.name] = [abi]

        try:
            return {
                abi_name: ContractCallHandler(contract=self, abis=abis)
                for abi_name, abis in view_methods.items()
            }
        except Exception as err:
            # NOTE: Must raise AttributeError for __attr__ method or will seg fault
            raise ApeAttributeError(str(err), base_err=err) from err

    @cached_property
    def _mutable_methods_(self) -> dict[str, ContractTransactionHandler]:
        mutable_methods: dict[str, list[MethodABI]] = dict()

        for abi in self.contract_type.mutable_methods:
            if abi.name in mutable_methods:
                mutable_methods[abi.name].append(abi)
            else:
                mutable_methods[abi.name] = [abi]

        try:
            return {
                abi_name: ContractTransactionHandler(contract=self, abis=abis)
                for abi_name, abis in mutable_methods.items()
            }
        except Exception as err:
            # NOTE: Must raise AttributeError for __attr__ method or will seg fault
            raise ApeAttributeError(str(err), base_err=err) from err

    def call_view_method(self, method_name: str, *args, **kwargs) -> Any:
        """
        Call a contract's view function directly using the method_name.
        This is helpful in the scenario where the contract has a
        method name matching an attribute of the
        :class:`~ape.api.address.BaseAddress` class, such as ``nonce``
        or ``balance``

        Args:
            method_name (str): The contract method name to be called
            *args: Contract method arguments.
            **kwargs: Transaction values, such as ``value`` or ``sender``

        Returns:
            Any: Output of smart contract view call.

        """

        if method_name in self._view_methods_:
            view_handler = self._view_methods_[method_name]
            output = view_handler(*args, **kwargs)
            return output

        elif method_name in self._mutable_methods_:
            handler = self._mutable_methods_[method_name].call
            output = handler(*args, **kwargs)
            return output

        else:
            # Didn't find anything that matches
            name = self.contract_type.name or "ContractType"
            raise ApeAttributeError(f"'{name}' has no attribute '{method_name}'.")

    def invoke_transaction(self, method_name: str, *args, **kwargs) -> "ReceiptAPI":
        """
        Call a contract's function directly using the method_name.
        This function is for non-view function's which may change
        contract state and will execute a transaction.
        This is helpful in the scenario where the contract has a
        method name matching an attribute of the
        :class:`~ape.api.address.BaseAddress` class, such as ``nonce``
        or ``balance``

        Args:
            method_name (str): The contract method name to be called
            *args: Contract method arguments.
            **kwargs: Transaction values, such as ``value`` or ``sender``

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`: Output of smart contract interaction.

        """

        if method_name in self._view_methods_:
            view_handler = self._view_methods_[method_name].transact
            output = view_handler(*args, **kwargs)
            return output

        elif method_name in self._mutable_methods_:
            handler = self._mutable_methods_[method_name]
            output = handler(*args, **kwargs)
            return output

        else:
            # Didn't find anything that matches
            name = self.contract_type.name or "ContractType"
            raise ApeAttributeError(f"'{name}' has no attribute '{method_name}'.")

    def get_event_by_signature(self, signature: str) -> ContractEvent:
        """
        Get an event by its signature. Most often, you can use the
        :meth:`~ape.contracts.base.ContractInstance.__getattr__`
        method on this class to access events. However, in the case
        when you have more than one event with the same name, such
        as the case where one event is coming from a base contract,
        you can use this method to access the respective events.

        Args:
            signature (str): The signature of the event.

        Returns:
            :class:`~ape.contracts.base.ContractEvent`
        """

        name_from_sig = signature.partition("(")[0].strip()
        options = self._events_.get(name_from_sig.strip(), [])

        err = ContractDataError(f"No event found with signature '{signature}'.")
        if not options:
            raise err

        for evt in options:
            if evt.abi.signature == signature:
                return evt

        raise err

    def get_error_by_signature(self, signature: str) -> type[CustomError]:
        """
        Get an error by its signature, similar to
        :meth:`~ape.contracts.ContractInstance.get_event_by_signature`.

        Args:
            signature (str): The signature of the error.

        Returns:
            :class:`~ape.exceptions.CustomError`
        """

        name_from_sig = signature.partition("(")[0].strip()
        options = self._errors_.get(name_from_sig, [])
        err = ContractDataError(f"No error found with signature '{signature}'.")
        if not options:
            raise err

        for contract_err in options:
            if contract_err.abi and contract_err.abi.signature == signature:
                return contract_err

        raise err

    @cached_property
    def _events_(self) -> dict[str, list[ContractEvent]]:
        events: dict[str, list[EventABI]] = {}

        for abi in self.contract_type.events:
            if abi.name in events:
                events[abi.name].append(abi)
            else:
                events[abi.name] = [abi]

        try:
            return {
                abi_name: [ContractEvent(contract=self, abi=abi) for abi in abi_list]
                for abi_name, abi_list in events.items()
            }
        except Exception as err:
            # NOTE: Must raise AttributeError for __attr__ method or will seg fault
            raise ApeAttributeError(str(err), base_err=err) from err

    @cached_property
    def _errors_(self) -> dict[str, list[type[CustomError]]]:
        abis: dict[str, list[ErrorABI]] = {}

        try:
            for abi in self.contract_type.errors:
                if abi.name in abis:
                    abis[abi.name].append(abi)
                else:
                    abis[abi.name] = [abi]

            # Check for prior error sub-class definitions for the same contract.
            prior_errors = self.chain_manager.contracts._get_errors(self.address)

            errors = {}
            for abi_name, abi_list in abis.items():
                errors_to_add = []
                for abi in abi_list:
                    error_type = None
                    for existing_cls in prior_errors:
                        if existing_cls.abi and existing_cls.abi.signature == abi.signature:
                            # Error class was previously defined by contract at same address.
                            error_type = existing_cls
                            break

                    if error_type is None:
                        # Error class is being defined for the first time.
                        error_type = self._create_custom_error_type(
                            abi, contract_address=self.address
                        )
                        self.chain_manager.contracts._cache_error(self.address, error_type)

                    errors_to_add.append(error_type)

                errors[abi_name] = errors_to_add

            return errors

        except Exception as err:
            # NOTE: Must raise AttributeError for __attr__ method or will seg fault
            raise ApeAttributeError(str(err), base_err=err) from err

    def __dir__(self) -> list[str]:
        """
        Display methods to IPython on ``c.[TAB]`` tab completion.

        Returns:
            list[str]
        """

        # NOTE: Type ignores because of this issue: https://github.com/python/typing/issues/1112
        #  They can be removed after next `mypy` release containing fix.
        values = [
            "contract_type",
            "txn_hash",
            self.decode_input.__name__,
            self.get_event_by_signature.__name__,
            self.invoke_transaction.__name__,
            self.call_view_method.__name__,
            ContractInstance.creation_metadata.fget.__name__,  # type: ignore[attr-defined]
        ]
        return list(
            set(self._base_dir_values).union(
                self._view_methods_, self._mutable_methods_, self._events_, values
            )
        )

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> Any:
        """
        Access a method, property, event, or error on the contract using ``.`` access.

        Usage example::

            result = contract.vote()  # Implies a method named "vote" exists on the contract.

        Args:
            attr_name (str): The name of the method or property to access.

        Returns:
            Any: The return value from the contract call, or a transaction receipt.
        """
        _assert_not_ipython_check(attr_name)
        if attr_name in set(super(BaseAddress, self).__dir__()):
            return super(BaseAddress, self).__getattribute__(attr_name)

        elif attr_name not in {
            *self._view_methods_,
            *self._mutable_methods_,
            *self._events_,
            *self._errors_,
        }:
            # Didn't find anything that matches
            # NOTE: `__getattr__` *must* raise `AttributeError`
            name = self.contract_type.name or ContractInstance.__name__
            raise ApeAttributeError(f"'{name}' has no attribute '{attr_name}'.")

        elif (
            int(attr_name in self._view_methods_)
            + int(attr_name in self._mutable_methods_)
            + int(attr_name in self._events_)
            + int(attr_name in self._errors_)
            > 1
        ):
            # ABI should not contain a mix of events, mutable and view methods that match
            # NOTE: `__getattr__` *must* raise `AttributeError`
            cls_name = getattr(type(self), "__name__", ContractInstance.__name__)
            raise ApeAttributeError(f"{cls_name} has corrupted ABI.")

        if attr_name in self._view_methods_:
            return self._view_methods_[attr_name]

        elif attr_name in self._mutable_methods_:
            return self._mutable_methods_[attr_name]

        elif attr_name in self._events_:
            evt_options = self._events_[attr_name]
            if len(evt_options) > 1:
                return ContractEventWrapper(evt_options)

            return evt_options[0]

        elif attr_name in self._errors_:
            err_options = self._errors_[attr_name]
            if len(err_options) > 1:
                raise ApeAttributeError(
                    f"Multiple errors named '{attr_name}' in '{self.contract_type.name}'.\n"
                    f"Use '{self.get_error_by_signature.__name__}' look-up."
                )

            return err_options[0]

        else:
            raise ApeAttributeError(
                f"No attribute '{attr_name}' found in contract '{self.address}'."
            )


class ContractContainer(ContractTypeWrapper, ExtraAttributesMixin):
    """
    A wrapper around the contract type that has access to the provider.
    When you import your contracts from the :class:`ape.managers.project.ProjectManager`, you
    are using this class.

    Usage example::

        from ape import project

        contract_container = project.MyContract  # Assuming there is a contract named "MyContract"
    """

    def __init__(self, contract_type: "ContractType") -> None:
        self.contract_type = contract_type

    @log_instead_of_fail(default="<ContractContainer>")
    def __repr__(self) -> str:
        return f"<{self.contract_type.name}>"

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> Any:
        """
        Access a contract error or event type via its ABI name using ``.`` access.

        Args:
            attr_name (str): The name of the event or error.

        Returns:
            :class:`~ape.types.ContractEvent` or a subclass of :class:`~ape.exceptions.CustomError`
            or any real attribute of the class.
        """
        return get_attribute_with_extras(self, attr_name)

    def __eq__(self, other):
        if not hasattr(other, "contract_type"):
            return NotImplemented

        return other.contract_type == self.contract_type

    def __ape_extra_attributes__(self) -> Iterator[ExtraModelAttributes]:
        yield ExtraModelAttributes(
            name="events",
            attributes=lambda x: (
                ContractEvent(contract=self, abi=self.contract_type.events[x])
                if x in self.contract_type.events
                else None
            ),
            include_getitem=True,
        )
        yield ExtraModelAttributes(
            name="errors",
            attributes=lambda x: (
                self._create_custom_error_type(self.contract_type.errors[x])
                if x in self.contract_type.errors
                else None
            ),
            include_getitem=True,
        )
        yield ExtraModelAttributes(
            name="contract_type",
            attributes=self.contract_type,
        )

    @property
    def deployments(self):
        """
        Contract deployments.

        Usage example::

            # Get the latest deployment
            my_contract = project.MyContract.deployments[-1]
        """

        return self.chain_manager.contracts.get_deployments(self)

    def at(
        self,
        address: "AddressType",
        txn_hash: str | HexBytes | None = None,
        fetch_from_explorer: bool = True,
        proxy_info: "ProxyInfoAPI | None" = None,
        detect_proxy: bool = True,
    ) -> ContractInstance:
        """
        Get a contract at the given address.

        Usage example::

            from ape import project

            my_contract = project.MyContract.at("0xAbC1230001112223334445566611855443322111")

        Args:
            address (str): The address to initialize a contract.
              **NOTE**: Things will not work as expected if the contract is not actually
              deployed to this address or if the contract at the given address has
              a different ABI than :attr:`~ape.contracts.ContractContainer.contract_type`.
            txn_hash (str | HexBytes): The hash of the transaction that deployed the
              contract, if available. Defaults to ``None``.
            fetch_from_explorer (bool): Set to ``False`` to avoid fetching from an explorer.
            proxy_info (:class:`~ape.api.networks.ProxyInfoAPI` | None): Proxy info object to set
              to avoid detection; defaults to ``None`` which will detect it if ``detect_proxy=True``.
            detect_proxy (bool): Set to ``False`` to avoid detecting missing proxy info.

        Returns:
            :class:`~ape.contracts.ContractInstance`
        """
        return self.chain_manager.contracts.instance_at(
            address,
            contract_type=self.contract_type,
            txn_hash=txn_hash,
            fetch_from_explorer=fetch_from_explorer,
            proxy_info=proxy_info,
            detect_proxy=detect_proxy,
        )

    @cached_property
    def constructor(self) -> ContractConstructor:
        return ContractConstructor(
            abi=self.contract_type.constructor,
            deployment_bytecode=self.contract_type.get_deployment_bytecode() or HexBytes(""),
        )

    def __call__(self, *args, **kwargs) -> "TransactionAPI":
        args_length = len(args)
        inputs_length = (
            len(self.constructor.abi.inputs)
            if self.constructor.abi and self.constructor.abi.inputs
            else 0
        )
        if inputs_length != args_length:
            raise ArgumentsLengthError(args_length, inputs=self.constructor.abi)

        return self.constructor.serialize_transaction(*args, **kwargs)

    def deploy(self, *args, publish: bool = False, **kwargs) -> ContractInstance:
        """
        Deploy a contract.

        Args:
            *args (Any): The contract's constructor arguments as Python types.
            publish (bool): Whether to also perform contract-verification.
              Defaults to ``False``.

        Returns:
            :class:`~ape.contracts.base.ContractInstance`
        """
        if isinstance(sender := kwargs.get("sender"), AccountAPI):
            possible_address = sender.get_deployment_address()
            styled_address = click.style(possible_address, bold=True)
            logger.info(f"Contract will be deployed at: {styled_address}")

        bytecode = self.contract_type.deployment_bytecode
        if not bytecode or bytecode.bytecode in (None, "", "0x"):
            raise MissingDeploymentBytecodeError(self.contract_type)

        txn = self(*args, **kwargs)
        private = kwargs.get("private", False)

        if kwargs.get("value") and not self.contract_type.constructor.is_payable:
            raise MethodNonPayableError("Sending funds to a non-payable constructor.")

        if "sender" in kwargs and hasattr(kwargs["sender"], "call"):
            # Handle account-related preparation if needed, such as signing
            receipt = self._cache_wrap(lambda: kwargs["sender"].call(txn, **kwargs))

        else:
            txn = self.provider.prepare_transaction(txn)
            receipt = self._cache_wrap(
                lambda: (
                    self.provider.send_private_transaction(txn)
                    if private
                    else self.provider.send_transaction(txn)
                )
            )

        address = receipt.contract_address
        if not address:
            raise ChainError(f"'{receipt.txn_hash}' did not create a contract.")

        styled_address = click.style(receipt.contract_address, bold=True)
        contract_name = self.contract_type.name or "<Unnamed Contract>"
        logger.success(f"Contract '{contract_name}' deployed to: {styled_address}")
        instance = ContractInstance.from_receipt(receipt, self.contract_type)
        self.chain_manager.contracts.cache_deployment(instance)

        if publish:
            self.local_project.deployments.track(instance)
            self.provider.network.publish_contract(address)

        instance.base_path = self.base_path or self.local_project.contracts_folder
        return instance

    def _cache_wrap(self, function: Callable) -> "ReceiptAPI":
        """
        A helper method to ensure a contract type is cached as early on
        as possible to help enrich errors from ``deploy()`` transactions
        as well produce nicer tracebacks for these errors. It also helps
        make assertions about these revert conditions in your tests.
        """
        try:
            return function()
        except ContractLogicError as err:
            if address := err.address:
                self.chain_manager.contracts[address] = self.contract_type
                err = err.with_ape_traceback()  # Re-try setting source traceback
                new_err = None
                try:
                    # Try enrichment again now that the contract type is cached.
                    new_err = self.compiler_manager.enrich_error(err)
                except Exception:
                    pass

                if new_err:
                    raise new_err from err

            raise  # The error after caching.

    def declare(self, *args, **kwargs) -> "ReceiptAPI":
        transaction = self.provider.network.ecosystem.encode_contract_blueprint(
            self.contract_type, *args, **kwargs
        )
        if "sender" in kwargs and hasattr(kwargs["sender"], "call"):
            return kwargs["sender"].call(transaction)

        receipt = self.provider.send_transaction(transaction)
        if receipt.contract_address:
            self.chain_manager.contracts.cache_blueprint(
                receipt.contract_address, self.contract_type
            )
        else:
            logger.debug("Failed to cache contract declaration: missing contract address.")

        return receipt


class ContractNamespace:
    """
    A class that bridges contract containers in a namespace.
    For example, if you have an interface structure like this::

        contracts:
          accounts:
            - interface.json
          mocks:
            - interface.json

    You can interact with them like this::

        account_interface = project.accounts.interface
        mock_interface = project.mocks.interface

    """

    def __init__(self, name: str, contracts: list[ContractContainer]):
        self.name = name
        self.contracts = contracts

    @log_instead_of_fail(default="<ContractNamespace>")
    def __repr__(self) -> str:
        return f"<{self.name}>"

    @only_raise_attribute_error
    def __getattr__(self, item: str) -> "ContractContainer | ContractNamespace":
        """
        Access the next contract container or namespace.

        Args:
            item (str): The name of the next node.

        Returns:
            :class:`~ape.contracts.base.ContractContainer` |
            :class:`~ape.contracts.base.ContractNamespace`]
        """
        _assert_not_ipython_check(item)

        def _get_name(cc: ContractContainer) -> str:
            return cc.contract_type.name or ""

        for contract in self.contracts:
            search_contract_name = _get_name(contract)
            search_name = (
                search_contract_name.replace(f"{self.name}.", "") if search_contract_name else None
            )
            if not search_name:
                continue

            elif search_name == item:
                return contract

            elif "." in search_name:
                next_node = search_name.partition(".")[0]
                if next_node != item:
                    continue

                subname = f"{self.name}.{next_node}"
                subcontracts = [c for c in self.contracts if _get_name(c).startswith(subname)]
                return ContractNamespace(subname, subcontracts)

        return self.__getattribute__(item)



from functools import cached_property
from typing import TYPE_CHECKING, Any

from ape.utils.basemodel import ManagerAccessMixin

if TYPE_CHECKING:
    from _pytest.config import Config as PytestConfig

    from ape.pytest.utils import Scope
    from ape.types.trace import ContractFunctionPath


def _get_config_exclusions(config) -> list["ContractFunctionPath"]:
    from ape.types.trace import ContractFunctionPath

    return [
        ContractFunctionPath(contract_name=x.contract_name, method_name=x.method_name)
        for x in config.exclude
    ]


class ConfigWrapper(ManagerAccessMixin):
    """
    A class aggregating settings choices from both the pytest command line
    and the ``ape-config.yaml`` file. Also serves as a wrapper around the
    Pytest config object for ease-of-use and code-sharing.
    """

    def __init__(self, pytest_config: "PytestConfig"):
        self.pytest_config = pytest_config
        if not self.verbosity:
            # Enable verbose output if stdout capture is disabled
            self.verbosity = self.pytest_config.getoption("capture") == "no"
        # else: user has already changes verbosity to an equal or higher level; avoid downgrading.

    @property
    def verbosity(self) -> int:
        return self.pytest_config.option.verbose

    @verbosity.setter
    def verbosity(self, value):
        self.pytest_config.option.verbose = value

    @cached_property
    def supports_tracing(self) -> bool:
        return self.provider.supports_tracing

    @cached_property
    def interactive(self) -> bool:
        return self.pytest_config.getoption("interactive")

    @cached_property
    def network(self) -> str:
        return self.pytest_config.getoption("network")

    @cached_property
    def isolation(self) -> bool:
        return not self.pytest_config.getoption("disable_isolation")

    @cached_property
    def disable_warnings(self) -> bool:
        return self.pytest_config.getoption("--disable-warnings")

    @cached_property
    def disconnect_providers_after(self) -> bool:
        return self.ape_test_config.disconnect_providers_after

    @cached_property
    def ape_test_config(self):
        return self.config_manager.get_config("test")

    @cached_property
    def track_gas(self) -> bool:
        return self.pytest_config.getoption("--gas") or self.ape_test_config.gas.show

    @cached_property
    def track_coverage(self) -> bool:
        return self.pytest_config.getoption("--coverage") or self.ape_test_config.coverage.track

    @property
    def xml_coverage(self) -> bool | dict:
        return self.ape_test_config.coverage.reports.xml

    @property
    def html_coverage(self) -> bool | dict:
        return self.ape_test_config.coverage.reports.html

    @cached_property
    def show_internal(self) -> bool:
        return self.pytest_config.getoption("--show-internal") or self.ape_test_config.show_internal

    @cached_property
    def enable_fixture_rebasing(self) -> bool:
        return self.ape_test_config.enable_fixture_rebasing

    @cached_property
    def gas_exclusions(self) -> list["ContractFunctionPath"]:
        """
        The combination of both CLI values and config values.
        """
        from ape.types.trace import ContractFunctionPath

        cli_value = self.pytest_config.getoption("--gas-exclude")
        exclusions = (
            [ContractFunctionPath.from_str(item) for item in cli_value.split(",")]
            if cli_value
            else []
        )
        paths = _get_config_exclusions(self.ape_test_config.gas)
        exclusions.extend(paths)
        return exclusions

    @cached_property
    def coverage_exclusions(self) -> list["ContractFunctionPath"]:
        return _get_config_exclusions(self.ape_test_config.coverage)

    def get_pytest_plugin(self, name: str) -> Any | None:
        if self.pytest_config.pluginmanager.has_plugin(name):
            return self.pytest_config.pluginmanager.get_plugin(name)

        return None

    def get_isolation(self, scope: "Scope"):
        if self.pytest_config.getoption("disable_isolation"):
            # Was disabled via command-line.
            return False

        # Check Ape's config.
        return self.ape_test_config.get_isolation(scope)



import re
from pathlib import Path

import pytest

from ape.api.transactions import ReceiptAPI
from ape.exceptions import (
    Abort,
    ContractLogicError,
    ContractNotFoundError,
    NetworkNotFoundError,
    TransactionError,
    UnknownSnapshotError,
    handle_ape_exception,
)
from ape.types.trace import SourceTraceback
from ape.utils.misc import LOCAL_NETWORK_NAME, ZERO_ADDRESS
from ape_ethereum.transactions import DynamicFeeTransaction, Receipt


@pytest.fixture(scope="module")
def failing_call():
    # A call (tx without sender)
    data = {
        "chainId": 1337,
        "to": "0x5FbDB2315678afecb367f032d93F642f64180aa3",
        "from": "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        "gas": 30029122,
        "value": 0,
        "data": "0xce50aa7d00000000000000000000000070997970c51812dc3a010c7d01b50e0d17dc79c80000000000000000000000000000000000000000000000000000000000000001",
        "type": 2,
        "maxFeePerGas": 875000000,
        "maxPriorityFeePerGas": 0,
        "accessList": [],
    }
    return DynamicFeeTransaction.model_validate(data)


class TestAbort:
    def test_shows_line_number(self):
        actual = str(Abort())
        expected = re.compile(r"Operation aborted in [\w<>.]*::[\w<>]* on line \d+\.")
        assert expected.match(actual)


class TestTransactionError:
    def test_receipt_is_subclass(self, vyper_contract_instance, owner):
        """
        Ensure TransactionError knows subclass Receipts are still receipts.
        (There was a bug once when it didn't, and that caused internal AttributeErrors).
        """

        class SubclassReceipt(Receipt):
            pass

        receipt = vyper_contract_instance.setNumber(123, sender=owner)
        receipt_data = {**receipt.model_dump(), "transaction": receipt.transaction}
        sub_receipt = SubclassReceipt.model_validate(receipt_data)

        err = TransactionError(txn=sub_receipt)
        assert isinstance(err.txn, ReceiptAPI)  # Same check used.

    def test_address(self, owner):
        err = TransactionError(contract_address=owner.address)
        assert err.address == owner.address

    def test_receiver_as_address(self, owner):
        tx = owner.transfer(owner, "1 wei")
        err = TransactionError(txn=tx)
        assert err.address == owner.address

    def test_deploy_address_as_address(self, owner, ethereum, project, zero_address):
        contract = project.VyperContract.deploy(629, sender=owner)

        receipt = contract.creation_metadata.receipt
        data = receipt.model_dump(exclude={"transaction"})
        # Show when receiver is zero_address, it still picks contract address.
        data["transaction"] = ethereum.create_transaction(receiver=zero_address)

        tx = Receipt.model_validate(data)
        assert tx.receiver == zero_address, "setup failed"

        err = TransactionError(txn=tx)
        assert err.address == contract.address

    def test_call_with_txn_and_not_source_tb(self, failing_call):
        """
        Simulating a failing-call, making sure it doesn't
        blow up if it doesn't get a source-tb.
        """
        err = TransactionError(txn=failing_call)
        assert err.source_traceback is None

    def test_call_with_source_tb_and_not_txn(self, mocker, project_with_contract):
        """
        Simulating a failing call, making sure the source-tb lines
        show up when a txn is NOT given.
        """
        # Using mocks for simplicity. Otherwise have to use a bunch of models from ethpm-types;
        # most of the stuff being mocked seems simple but is calculated from AST-Nodes and such.
        src_path = "path/to/VyperFile.vy"
        mock_tb = mocker.MagicMock()
        mock_exec = mocker.MagicMock()
        mock_exec.depth = 1
        mock_exec.source_path = src_path
        mock_exec.begin_lineno = 5
        mock_exec.end_lineno = 5
        mock_closure = mocker.MagicMock()
        mock_closure.name = "setNumber"
        mock_exec.closure = mock_closure
        mock_tb.__getitem__.return_value = mock_exec
        mock_tb.__len__.return_value = 1
        mock_tb.return_value = mock_tb
        err = TransactionError(
            source_traceback=mock_tb, project=project_with_contract, set_ape_traceback=True
        )

        # Have to raise for sys.exc_info() to be available.
        try:
            raise err
        except Exception:
            pass

        def assert_ape_traceback(err_arg):
            assert err_arg.__traceback__ is not None
            # The Vyper-frame gets injected at tb_next.
            assert err_arg.__traceback__.tb_next is not None
            actual = str(err_arg.__traceback__.tb_next.tb_frame)
            assert src_path in actual

        assert_ape_traceback(err)

        err2 = TransactionError(
            source_traceback=mock_tb,
            project=project_with_contract,
            set_ape_traceback=False,
        )
        try:
            raise err2
        except Exception:
            pass

        # No Ape frames are here.
        if err2.__traceback__:
            assert err2.__traceback__.tb_next is None

        err3 = ContractLogicError(source_traceback=mock_tb, project=project_with_contract)
        try:
            raise err3
        except Exception:
            pass

        assert_ape_traceback(err3)

    def test_source_traceback_from_txn(self, owner):
        """
        Was not given a source-traceback but showing we can deduce one from
        the given transaction.
        """
        tx = owner.transfer(owner, 0)
        err = TransactionError(txn=tx)
        _ = err.source_traceback
        assert err._attempted_source_traceback


class TestNetworkNotFoundError:
    def test_close_match(self):
        net = "sepolai"
        error = NetworkNotFoundError(net, ecosystem="ethereum", options=("sepolia",))
        actual = str(error)
        expected = f"No network in 'ethereum' named '{net}'. Did you mean 'sepolia'?"
        assert actual == expected

    def test_no_close_matches(self):
        net = "madeup"
        error = NetworkNotFoundError(net, ecosystem="ethereum", options=("sepolia",))
        actual = str(error)
        expected = f"No network in 'ethereum' named '{net}'. Options:\nsepolia"
        assert actual == expected

    def test_ecosystem_no_network_options(self):
        net = "madeup"
        error = NetworkNotFoundError(net, ecosystem="ethereum", options=())
        actual = str(error)
        expected = "'ethereum' has no networks."
        assert actual == expected

    def test_no_options(self):
        net = "madeup"
        error = NetworkNotFoundError(net, options=())
        actual = str(error)
        expected = "No networks found."
        assert actual == expected


def test_handle_ape_exception_hides_home_dir(mocker):
    base_paths = ["this/is/base/path"]
    mock_tb_get = mocker.patch("ape.exceptions._get_relevant_frames")
    tb_str = f"""
  File "{Path.home()}/this/is/base/path/myfile.vy", line 18, in setNumber2
    setNumber(5)
""".lstrip()
    mock_tb_get.return_value = [tb_str]
    print_watcher = mocker.patch("ape.exceptions.rich_print")
    ape_exception = ContractLogicError()
    assert handle_ape_exception(ape_exception, base_paths) is True
    actual = print_watcher.call_args[0][0]

    # We expect the same only the home dir has been hidden.
    expected = "\n" + tb_str.replace(str(Path.home()), "$HOME")
    assert actual == expected


class TestContractLogicError:
    @pytest.mark.parametrize("revert_message", (None, ""))
    def test_message_uses_revert_type_when_no_revert_message(self, mocker, revert_message):
        class TB(SourceTraceback):
            @property
            def revert_type(self) -> str | None:
                return "CUSTOM_ERROR"

        tb = TB([{"statements": [], "closure": {"name": "fn"}, "depth": 0}])  # type: ignore
        error = ContractLogicError(revert_message=revert_message, source_traceback=tb)
        actual = error.message
        expected = "CUSTOM_ERROR"
        assert actual == expected


class TestContractNotFoundError:
    def test_local_network(self):
        """
        Testing we are NOT mentioning explorer plugins
        for the local-network, as 99.9% of the time it is
        confusing.
        """
        err = ContractNotFoundError(ZERO_ADDRESS, False, f"ethereum:{LOCAL_NETWORK_NAME}:test")
        assert str(err) == f"Failed to get contract type for address '{ZERO_ADDRESS}'."

    def test_fork_network(self):
        err = ContractNotFoundError(ZERO_ADDRESS, False, "ethereum:sepolia-fork:test")
        assert str(err) == (
            f"Failed to get contract type for address '{ZERO_ADDRESS}'. "
            "Current network 'ethereum:sepolia-fork:test' has no associated explorer plugin. "
            "Try installing an explorer plugin using \x1b[32mape plugins install etherscan"
            "\x1b[0m, or using a network with explorer support."
        )

    def test_revert_message_setter(self):
        err = ContractLogicError(revert_message="test message 1")
        err.revert_message = "test message 2"
        assert str(err) == "test message 2"
        assert err.message == "test message 2"
        assert err.revert_message == "test message 2"


class TestUnknownSnapshotError:
    def test_bytes(self):
        snapshot_id = b"asdfasdfasdf"
        err = UnknownSnapshotError(snapshot_id)
        assert str(err) == "Unknown snapshot ID '0x6173..6466'."

    @pytest.mark.parametrize("snapshot_id", (123, "123"))
    def test_not_bytes(self, snapshot_id):
        err = UnknownSnapshotError(snapshot_id)
        assert str(err) == "Unknown snapshot ID '123'."



from importlib import import_module

from ape import plugins


@plugins.register(plugins.CompilerPlugin)
def register_compiler():
    compiler = import_module("ape_pm.compiler")
    return (".json",), compiler.InterfaceCompiler


@plugins.register(plugins.DependencyPlugin)
def dependencies():
    _dependencies = import_module("ape_pm.dependency")
    yield "github", _dependencies.GithubDependency
    yield "local", _dependencies.LocalDependency
    yield "npm", _dependencies.NpmDependency
    yield ("python", "pypi", "site_package"), _dependencies.PythonDependency


@plugins.register(plugins.ProjectPlugin)
def projects():
    _projects = import_module("ape_pm.project")
    yield _projects.BrownieProject
    yield _projects.FoundryProject


def __getattr__(name: str):
    if name in ("BrownieProject", "FoundryProject"):
        module = import_module("ape_pm.project")
    elif name in ("GithubDependency", "LocalDependency", "NpmDependency", "PythonDependency"):
        module = import_module("ape_pm.dependency")
    else:
        module = import_module("ape_pm.compiler")

    return getattr(module, name)


__all__ = [
    "BrownieProject",
    "FoundryProject",
    "GithubDependency",
    "InterfaceCompiler",
    "LocalDependency",
    "NpmDependency",
    "PythonDependency",
]



from typing import TYPE_CHECKING

from .pluggy_patch import PluginType, hookspec

if TYPE_CHECKING:
    from ape.api.accounts import AccountAPI, AccountContainerAPI


class AccountPlugin(PluginType):
    """
    An account-related plugin. The plugin must register both
    an :class:`ape.api.accounts.AccountContainerAPI` as well as an
    :class:`ape.api.accounts.AccountAPI`.
    """

    @hookspec
    def account_types(  # type: ignore[empty-body]
        self,
    ) -> tuple[type["AccountContainerAPI"], type["AccountAPI"]]:
        """
        A hook for returning a tuple of an account container and an account type.
        Each account-base plugin defines and returns their own types here.

        Usage example::

            @plugins.register(plugins.AccountPlugin)
            def account_types():
                return AccountContainer, KeyfileAccount


        Returns:
            tuple[type[:class:`~ape.api.accounts.AccountContainerAPI`],
            type[:class:`~ape.api.accounts.AccountAPI`]]
        """



import pytest

from ape.types.signatures import MessageSignature, SignableMessage, TransactionSignature


@pytest.fixture
def signable_message():
    version = b"E"
    header = b"thereum Signed Message:\n32"
    body = (
        b"\x86\x05\x99\xc6\xfa\x0f\x05|po(\x1f\xe3\x84\xc0\x0f"
        b"\x13\xb2\xa6\x91\xa3\xb8\x90\x01\xc0z\xa8\x17\xbe'\xf3\x13"
    )
    return SignableMessage(version=version, header=header, body=body)


@pytest.fixture
def signature(owner, signable_message):
    return owner.sign_message(signable_message)


def test_signature_repr():
    signature = TransactionSignature(v=0, r=b"123", s=b"456")
    assert repr(signature) == "<TransactionSignature v=0 r=0x313233 s=0x343536>"


def test_signable_message_repr(signable_message):
    actual = repr(signable_message)
    expected_version = "E"
    expected_header = "thereum Signed Message:\n32"
    expected_body = "0x860599c6fa0f057c706f281fe384c00f13b2a691a3b89001c07aa817be27f313"
    expected = (
        f'SignableMessage(version="{expected_version}", header="{expected_header}", '
        f'body="{expected_body}")'
    )

    assert actual == expected


def test_signature_from_rsv_and_vrs(signature):
    rsv = signature.encode_rsv()
    vrs = signature.encode_vrs()

    # NOTE: Type declaring for sake of ensuring
    #   type-checking works since class-method is
    #   defined in base-class.
    from_rsv: MessageSignature = signature.from_rsv(rsv)
    from_vrs: MessageSignature = signature.from_vrs(vrs)
    assert from_rsv == from_vrs == signature


def test_signable_message_module(signable_message):
    """
    At the time of writing this, SignableMessage is a borrowed
    construct from `eth_account.messages`. We are changing the module
    manually so we are testing that it shows Ape's now.
    """
    actual = signable_message.__module__
    expected = "ape.types.signatures"
    assert actual == expected

    # Also show the 404-causing line in the __doc__ was fixed.
    assert "EIP-191_" not in signable_message.__doc__
    assert "EIP-191" in signable_message.__doc__



from typing import Any

import pytest

from ape.api.convert import ConvertibleAPI
from ape.types.address import AddressType


@pytest.fixture(scope="module")
def custom_type(accounts):
    class MyAccountWrapper(ConvertibleAPI):
        def __init__(self, acct):
            self.acct = acct

        def is_convertible(self, to_type: type) -> bool:
            return to_type is AddressType

        def convert_to(self, to_type: type) -> Any:
            if to_type is AddressType:
                return self.acct.address

            raise NotImplementedError()

    return MyAccountWrapper(accounts[0])


def test_convert(custom_type, conversion_manager, accounts):
    """
    You can use the regular conversion manager to convert the custom type.
    """
    actual = conversion_manager.convert(custom_type, AddressType)
    assert actual == accounts[0].address



from abc import abstractmethod
from collections.abc import Iterable, Iterator
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from ape.exceptions import APINotImplementedError, ContractLogicError
from ape.utils.basemodel import BaseInterfaceModel
from ape.utils.misc import log_instead_of_fail, raises_not_implemented

if TYPE_CHECKING:
    from eth_pydantic_types import HexBytes
    from ethpm_types import ContractType
    from ethpm_types.source import Content, ContractSource
    from packaging.version import Version

    from ape.api.config import PluginConfig
    from ape.api.trace import TraceAPI
    from ape.managers.project import ProjectManager
    from ape.types.coverage import ContractSourceCoverage
    from ape.types.trace import SourceTraceback


class CompilerAPI(BaseInterfaceModel):
    """
    Compiler plugins, such as for languages like
    `Solidity <https://docs.soliditylang.org/en/v0.8.11/>`__ or
    `Vyper <https://vyper.readthedocs.io/en/stable/>`__, implement this API.

    See the repository for the `ape-solidity <https://github.com/ApeWorX/ape-solidity>`__ plugin or
    the `ape-vyper <https://github.com/ApeWorX/ape-vyper>`__ plugin as example implementations of
    this API.
    """

    compiler_settings: dict = {}
    """
    Adhoc compiler settings.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of the compiler.
        """

    def get_config(self, project: "ProjectManager | None" = None) -> "PluginConfig":
        """
        The combination of settings from ``ape-config.yaml`` and ``.compiler_settings``.

        Args:
            project ("ProjectManager | None"): Optionally provide
              the project containing the base paths and full source set. Defaults to the local
              project. Dependencies will change this value to their respective projects.

        Returns:
            :class:`~ape.api.config.PluginConfig`
        """
        pm = project or self.local_project
        config = pm.config.get_config(self.name)
        data = {**config.model_dump(mode="json", by_alias=True), **self.compiler_settings}
        return config.model_validate(data)

    @raises_not_implemented
    def get_versions(self, all_paths: Iterable[Path]) -> set[str]:  # type: ignore[empty-body]
        """
        Retrieve the set of available compiler versions for this plugin to compile ``all_paths``.

        Args:
            all_paths (Iterable[pathlib.Path]): The list of paths.

        Returns:
            set[str]: A set of available compiler versions.
        """

    @raises_not_implemented
    def get_compiler_settings(  # type: ignore[empty-body]
        self,
        contract_filepaths: Iterable[Path],
        project: "ProjectManager | None" = None,
        **overrides,
    ) -> dict["Version", dict]:
        """
        Get a mapping of the settings that would be used to compile each of the sources
        by the compiler version number.

        Args:
            contract_filepaths (Iterable[pathlib.Path]): The list of paths.
            project ("ProjectManager | None"): Optionally provide
              the project containing the base paths and full source set. Defaults to the local
              project. Dependencies will change this value to their respective projects.
            **overrides: Settings overrides.

        Returns:
            dict[Version, dict]: A dict of compiler settings by compiler version.
        """

    @abstractmethod
    def compile(
        self,
        contract_filepaths: Iterable[Path],
        project: "ProjectManager | None",
        settings: dict | None = None,
    ) -> Iterator["ContractType"]:
        """
        Compile the given source files. All compiler plugins must implement this function.

        Args:
            contract_filepaths (Iterable[pathlib.Path]): A list of source file paths to compile.
            project ("ProjectManager | None"): Optionally provide
              the project containing the base paths and full source set. Defaults to the local
              project. Dependencies will change this value to their respective projects.
            settings (dict | None): Adhoc compiler settings.

        Returns:
            list[:class:`~ape.type.contract.ContractType`]
        """

    @raises_not_implemented
    def compile_code(  # type: ignore[empty-body]
        self,
        code: str,
        project: "ProjectManager | None",
        settings: dict | None = None,
        **kwargs,
    ) -> "ContractType":
        """
        Compile a program.

        Args:
            code (str): The code to compile.
            project ("ProjectManager | None"): Optionally provide
              the project containing the base paths and full source set. Defaults to the local
              project. Dependencies will change this value to their respective projects.
            settings (Optional[dict]): Adhoc compiler settings.
            **kwargs: Additional overrides for the ``ethpm_types.ContractType`` model.

        Returns:
            ``ContractType``: A compiled contract artifact.
        """

    @raises_not_implemented
    def get_imports(  # type: ignore[empty-body]
        self, contract_filepaths: Iterable[Path], project: "ProjectManager | None"
    ) -> dict[str, list[str]]:
        """
        Returns a list of imports as source_ids for each contract's source_id in a given
        compiler.

        Args:
            contract_filepaths (Iterable[pathlib.Path]): A list of source file paths to compile.
            project ("ProjectManager | None"): Optionally provide
              the project containing the base paths and full source set. Defaults to the local
              project. Dependencies will change this value to their respective projects.

        Returns:
            dict[str, list[str]]: A dictionary like ``{source_id: [import_source_id, ...], ...}``
        """

    @raises_not_implemented
    def get_version_map(  # type: ignore[empty-body]
        self,
        contract_filepaths: Iterable[Path],
        project: "ProjectManager | None" = None,
    ) -> dict["Version", set[Path]]:
        """
        Get a map of versions to source paths.

        Args:
            contract_filepaths (Iterable[Path]): Input source paths. Defaults to all source paths
              per compiler.
            project ("ProjectManager | None"): Optionally provide
              the project containing the base paths and full source set. Defaults to the local
              project. Dependencies will change this value to their respective projects.

        Returns:
            dict[Version, set[Path]]
        """

    @log_instead_of_fail(default="<CompilerAPI>")
    def __repr__(self) -> str:
        cls_name = getattr(type(self), "__name__", CompilerAPI.__name__)
        return f"<{cls_name} {self.name}>"

    def __str__(self) -> str:
        return self.name

    @cached_property
    def supports_source_tracing(self) -> bool:
        """
        Returns ``True`` if this compiler is able to provider a source
        traceback for a given trace.
        """
        try:
            self.trace_source(None, None, None)  # type: ignore
        except APINotImplementedError:
            return False
        except Exception:
            # Task failed successfully.
            return True

        return True

    def enrich_error(self, err: ContractLogicError) -> ContractLogicError:
        """
        Enrich a contract logic error using compiler information, such as
        known PC locations for compiler runtime errors.

        Args:
            err (:class:`~ape.exceptions.ContractLogicError`): The exception
              to enrich.

        Returns:
            :class:`~ape.exceptions.ContractLogicError`: The enriched exception.
        """

        return err

    @raises_not_implemented
    def trace_source(  # type: ignore[empty-body]
        self, contract_source: "ContractSource", trace: "TraceAPI", calldata: "HexBytes"
    ) -> "SourceTraceback":
        """
        Get a source-traceback for the given contract type.
        The source traceback object contains all the control paths taken in the transaction.
        When available, source-code location information is accessible from the object.

        Args:
            contract_source (``ContractSource``): A contract type with a local-source that was
              compiled by this compiler.
            trace (:class:`~ape.api.trace.TraceAPI`]): The resulting trace from executing a
              function defined in the given contract type.
            calldata (``HexBytes``): Calldata passed to the top-level call.

        Returns:
            :class:`~ape.types.trace.SourceTraceback`
        """

    @raises_not_implemented
    def flatten_contract(  # type: ignore[empty-body]
        self, path: Path, project: "ProjectManager | None" = None, **kwargs
    ) -> "Content":
        """
        Get the content of a flattened contract via its source path.
        Plugin implementations handle import resolution, SPDX de-duplication,
        and anything else needed.

        Args:
            path (``pathlib.Path``): The source path of the contract.
            project ("ProjectManager | None"): Optionally provide
              the project containing the base paths and full source set. Defaults to the local
              project. Dependencies will change this value to their respective projects.
            **kwargs (Any): Additional compiler-specific settings. See specific
              compiler plugins when applicable.

        Returns:
            ``ethpm_types.source.Content``: The flattened contract content.
        """

    @raises_not_implemented
    def init_coverage_profile(
        self, source_coverage: "ContractSourceCoverage", contract_source: "ContractSource"
    ):  # type: ignore[empty-body]
        """
        Initialize an empty report for the given source ID. Modifies the given source
        coverage in-place.

        Args:
            source_coverage (:class:`~ape.types.coverage.SourceCoverage`): The source
              to generate an empty coverage profile for.
            contract_source (``ethpm_types.source.ContractSource``): The contract with
              source content.
        """



import shutil

from ape.exceptions import ProviderNotConnectedError


def test_networks_context_does_not_disconnect(networks):
    try:
        assert networks.provider.web3
    except ProviderNotConnectedError:
        # Show more info for debugging purposes
        context = networks.ethereum.local.use_provider("test")
        error_output = f"""
        ape: {shutil.which("ape")},
        connected_providers: {", ".join([pid for pid in context.connected_providers.keys()])},
        """
        assert False, f"Provider lost connection!\n{error_output}"

    context = networks.ethereum.local.use_provider("test")

    with context as provider:
        assert provider

    # Ensure we are still connected
    assert networks.provider.web3



from ape.cli.arguments import (
    contract_file_paths_argument,
    existing_alias_argument,
    non_existing_alias_argument,
)
from ape.cli.choices import (
    AccountAliasPromptChoice,
    Alias,
    LazyChoice,
    NetworkChoice,
    OutputFormat,
    PromptChoice,
    output_format_choice,
    select_account,
)
from ape.cli.commands import ConnectedProviderCommand
from ape.cli.options import (
    ApeCliContextObject,
    NetworkOption,
    account_option,
    ape_cli_context,
    config_override_option,
    contract_option,
    incompatible_with,
    network_option,
    output_format_option,
    project_option,
    provider_settings_option,
    skip_confirmation_option,
    verbosity_option,
)
from ape.cli.paramtype import JSON, Noop, Path

__all__ = [
    "JSON",
    "AccountAliasPromptChoice",
    "Alias",
    "ApeCliContextObject",
    "ConnectedProviderCommand",
    "LazyChoice",
    "NetworkChoice",
    "NetworkOption",
    "Noop",
    "OutputFormat",
    "Path",
    "PromptChoice",
    "account_option",
    "ape_cli_context",
    "config_override_option",
    "contract_file_paths_argument",
    "contract_option",
    "existing_alias_argument",
    "incompatible_with",
    "network_option",
    "non_existing_alias_argument",
    "output_format_choice",
    "output_format_option",
    "project_option",
    "provider_settings_option",
    "select_account",
    "skip_confirmation_option",
    "verbosity_option",
]



from collections.abc import Iterator
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING

from eth_pydantic_types import HexBytes
from ethpm_types import ASTNode, BaseModel
from ethpm_types.source import (
    Closure,
    Content,
    ContractSource,
    Function,
    SourceStatement,
    Statement,
)
from pydantic import RootModel
from pydantic.dataclasses import dataclass

from ape.utils.misc import log_instead_of_fail

if TYPE_CHECKING:
    from ethpm_types.ast import SourceLocation

    from ape.api.trace import TraceAPI


GasReport = dict[str, dict[str, list[int]]]
"""
A gas report in Ape.
"""


class ControlFlow(BaseModel):
    """
    A collection of linear source nodes up until a jump.
    """

    statements: list[Statement]
    """
    The source node statements.
    """

    closure: Closure
    """
    The defining closure, such as a function or module, of the code sequence.
    """

    source_path: Path | None = None
    """
    The path to the local contract file.
    Only exists when is from a local contract.
    """

    depth: int
    """
    The depth at which this flow was executed,
    where 1 is the first calling function.
    """

    def __str__(self) -> str:
        return f"{self.source_header}\n{self.format()}"

    @log_instead_of_fail(default="<ControlFlow>")
    def __repr__(self) -> str:
        source_name = f" {self.source_path.name} " if self.source_path is not None else " "
        representation = f"<control path,{source_name}{self.closure.name}"

        if len(self.statements) > 0:
            representation = f"{representation} num_statements={len(self.statements)}"

        if self.begin_lineno is None:
            return f"{representation}>"

        else:
            # Include line number info.
            end_lineno = self.end_lineno or self.begin_lineno
            line_range = (
                f"line {self.begin_lineno}"
                if self.begin_lineno == end_lineno
                else f"lines {self.begin_lineno}-{end_lineno}"
            )
            return f"{representation}, {line_range}>"

    def __getitem__(self, idx: int) -> Statement:
        try:
            return self.statements[idx]
        except IndexError as err:
            raise IndexError(f"Statement index '{idx}' out of range.") from err

    def __len__(self) -> int:
        return len(self.statements)

    @property
    def source_statements(self) -> list[SourceStatement]:
        """
        All statements coming directly from a contract's source.
        Excludes implicit-compiler statements.
        """
        return [x for x in self.statements if isinstance(x, SourceStatement)]

    @property
    def begin_lineno(self) -> int | None:
        """
        The first line number in the sequence.
        """
        stmts = self.source_statements
        return stmts[0].begin_lineno if stmts else None

    @property
    def ws_begin_lineno(self) -> int | None:
        """
        The first line number in the sequence, including whitespace.
        """
        stmts = self.source_statements
        return stmts[0].ws_begin_lineno if stmts else None

    @property
    def line_numbers(self) -> list[int]:
        """
        The list of all line numbers as part of this node.
        """

        if self.begin_lineno is None:
            return []

        elif self.end_lineno is None:
            return [self.begin_lineno]

        return list(range(self.begin_lineno, self.end_lineno + 1))

    @property
    def content(self) -> Content:
        result: dict[int, str] = {}
        for node in self.source_statements:
            result = {**result, **node.content.root}

        return Content(root=result)

    @property
    def source_header(self) -> str:
        result = ""
        if self.source_path is not None:
            result += f"File {self.source_path}, in "

        result += f"{self.closure.name}"
        return result.strip()

    @property
    def end_lineno(self) -> int | None:
        """
        The last line number.
        """
        stmts = self.source_statements
        return stmts[-1].end_lineno if stmts else None

    @property
    def pcs(self) -> set[int]:
        full_set: set[int] = set()
        for stmt in self.statements:
            full_set |= stmt.pcs

        return full_set

    def extend(
        self,
        location: "SourceLocation",
        pcs: set[int] | None = None,
        ws_start: int | None = None,
    ):
        """
        Extend this node's content with other content that follows it directly.

        Raises:
            ValueError: When there is a gap in content.

        Args:
            location (SourceLocation): The location of the content, in the form
              (lineno, col_offset, end_lineno, end_coloffset).
            pcs (set[int] | None): The PC values of the statements.
            ws_start (int | None): Optionally provide a white-space starting point
              to back-fill.
        """

        pcs = pcs or set()
        if ws_start is not None and ws_start > location[0]:
            # No new lines.
            return

        function = self.closure
        if not isinstance(function, Function):
            # No source code supported for closure type.
            return

        # NOTE: Use non-ws prepending location to fetch AST nodes.
        asts = function.get_content_asts(location)
        if not asts:
            return

        location = (
            (ws_start, location[1], location[2], location[3]) if ws_start is not None else location
        )
        content = function.get_content(location)
        start = (
            max(location[0], self.end_lineno + 1) if self.end_lineno is not None else location[0]
        )
        end = location[0] + len(content) - 1
        if end < start:
            # No new lines.
            return

        elif start - end > 1:
            raise ValueError(
                "Cannot extend when gap in lines > 1. "
                "If because of whitespace lines, must include it the given content."
            )

        content_start = len(content) - (end - start) - 1
        new_lines = {no: ln.rstrip() for no, ln in content.items() if no >= content_start}
        if new_lines:
            # Add the next statement in this sequence.
            content = Content(root=new_lines)
            statement = SourceStatement(asts=asts, content=content, pcs=pcs)
            self.statements.append(statement)

        else:
            # Add ASTs to latest statement.
            self.source_statements[-1].asts.extend(asts)
            for pc in pcs:
                self.source_statements[-1].pcs.add(pc)

    def format(self, use_arrow: bool = True) -> str:
        """
        Format this trace node into a string presentable to the user.
        """

        # NOTE: Only show last 2 statements.
        relevant_stmts = self.statements[-2:]
        content = ""
        end_lineno = self.content.end_lineno

        for stmt in relevant_stmts:
            for lineno, line in getattr(stmt, "content", {}).items():
                if not content and not line.strip():
                    # Prevent starting on whitespace.
                    continue

                if content:
                    # Add newline to carry over from last.
                    content = f"{content.rstrip()}\n"

                space = "       " if lineno < end_lineno or not use_arrow else "  -->  "
                content = f"{content}{space}{lineno} {line}"

        return content

    @property
    def next_statement(self) -> SourceStatement | None:
        """
        Returns the next statement that _would_ execute
        if the program were to progress to the next line.
        """

        # Check for more statements that _could_ execute.
        if not self.source_statements:
            return None

        last_stmt = self.source_statements[-1]
        function = self.closure
        if not isinstance(function, Function):
            return None

        rest_asts = [a for a in function.ast.children if a.lineno > last_stmt.end_lineno]
        if not rest_asts:
            # At the end of a function.
            return None

        # Filter out to only the ASTs for the next statement.
        next_stmt_start = min(rest_asts, key=lambda x: x.lineno).lineno
        next_stmt_asts = [a for a in rest_asts if a.lineno == next_stmt_start]
        content_dict = {}
        for ast in next_stmt_asts:
            sub_content = function.get_content(ast.line_numbers)
            content_dict = {**sub_content.root}

        if not content_dict:
            return None

        sorted_dict = {k: content_dict[k] for k in sorted(content_dict)}
        content = Content(root=sorted_dict)
        return SourceStatement(asts=next_stmt_asts, content=content)


class SourceTraceback(RootModel[list[ControlFlow]]):
    """
    A full execution traceback including source code.
    """

    @classmethod
    def create(cls, contract_source: ContractSource, trace: "TraceAPI", data: HexBytes | str):
        # Use the trace as a 'ManagerAccessMixin'.
        compilers = trace.compiler_manager
        source_id = contract_source.source_id
        ext = f".{source_id.split('.')[-1]}"
        if ext not in compilers.registered_compilers:
            return cls.model_validate([])

        compiler = compilers.registered_compilers[ext]
        try:
            return compiler.trace_source(contract_source, trace, HexBytes(data))
        except NotImplementedError:
            return cls.model_validate([])

    def __str__(self) -> str:
        return self.format()

    @log_instead_of_fail(default="<SourceTraceback>")
    def __repr__(self) -> str:
        return f"<ape.types.SourceTraceback control_paths={len(self.root)}>"

    def __len__(self) -> int:
        return len(self.root)

    def __iter__(self) -> Iterator[ControlFlow]:  # type: ignore[override]
        yield from self.root

    def __getitem__(self, idx: int) -> ControlFlow:
        try:
            return self.root[idx]
        except IndexError as err:
            raise IndexError(f"Control flow index '{idx}' out of range.") from err

    def __setitem__(self, key, value):
        return self.root.__setitem__(key, value)

    @property
    def revert_type(self) -> str | None:
        """
        The revert type, such as a builtin-error code or a user dev-message,
        if there is one.
        """
        return self.statements[-1].type if self.statements[-1].type != "source" else None

    def append(self, __object) -> None:
        """
        Append the given control flow to this one.
        """
        self.root.append(__object)

    def extend(self, __iterable) -> None:
        """
        Append all the control flows from the given traceback to this one.
        """
        if not isinstance(__iterable, SourceTraceback):
            raise TypeError("Can only extend another traceback object.")

        self.root.extend(__iterable.root)

    @property
    def last(self) -> ControlFlow | None:
        """
        The last control flow in the traceback, if there is one.
        """
        return self.root[-1] if len(self.root) else None

    @property
    def execution(self) -> list[ControlFlow]:
        """
        All the control flows in order. Each set of statements in
        a control flow is separated by a jump.
        """
        return list(self.root)

    @property
    def statements(self) -> list[Statement]:
        """
        All statements from each control flow.
        """
        return list(chain(*[x.statements for x in self.root]))

    @property
    def source_statements(self) -> list[SourceStatement]:
        """
        All source statements from each control flow.
        """
        return list(chain(*[x.source_statements for x in self.root]))

    def format(self) -> str:
        """
        Get a formatted traceback string for displaying to users.
        """
        if not len(self.root):
            # No calls.
            return ""

        header = "Traceback (most recent call last)"
        indent = "  "
        last_depth = None
        segments: list[str] = []
        for control_flow in reversed(self.root):
            if last_depth is None or control_flow.depth == last_depth - 1:
                if control_flow.depth == 0 and len(segments) >= 1:
                    # Ignore 0-layer segments if source code was hit
                    continue

                last_depth = control_flow.depth
                content_str = control_flow.format()
                if not content_str.strip():
                    continue

                segment = f"{indent}{control_flow.source_header}\n{control_flow.format()}"

                # Try to include next statement for display purposes.
                next_stmt = control_flow.next_statement
                if next_stmt:
                    if (
                        next_stmt.begin_lineno is not None
                        and control_flow.end_lineno is not None
                        and next_stmt.begin_lineno > control_flow.end_lineno + 1
                    ):
                        # Include whitespace.
                        for ws_no in range(control_flow.end_lineno + 1, next_stmt.begin_lineno):
                            function = control_flow.closure
                            if not isinstance(function, Function):
                                continue

                            ws = function.content[ws_no]
                            segment = f"{segment}\n       {ws_no} {ws}".rstrip()

                    for no, line in next_stmt.content.items():
                        segment = f"{segment}\n       {no} {line}".rstrip()

                segments.append(segment)

        builder = ""
        for idx, segment in enumerate(reversed(segments)):
            builder = f"{builder}\n{segment}"

            if idx < len(segments) - 1:
                builder = f"{builder}\n"

        return f"{header}{builder}"

    def add_jump(
        self,
        location: "SourceLocation",
        function: Function,
        depth: int,
        pcs: set[int] | None = None,
        source_path: Path | None = None,
    ):
        """
        Add an execution sequence from a jump.

        Args:
            location (``SourceLocation``): The location to add.
            function (``Function``): The function executing.
            source_path (Path | None): The path of the source file.
            depth (int): The depth of the function call in the call tree.
            pcs (set[int] | None): The program counter values.
            source_path (Path | None): The path of the source file.
        """

        asts = function.get_content_asts(location)
        content = function.get_content(location)
        if not asts or not content:
            return

        pcs = pcs or set()
        Statement.model_rebuild()
        ControlFlow.model_rebuild()
        self._add(asts, content, pcs, function, depth, source_path=source_path)

    def extend_last(self, location: "SourceLocation", pcs: set[int] | None = None):
        """
        Extend the last node with more content.

        Args:
            location (``SourceLocation``): The location of the new content.
            pcs (set[int] | None): The PC values to add on.
        """

        if not self.last:
            raise ValueError(
                "`progress()` should only be called when "
                "there is at least 1 ongoing execution trail."
            )

        start = (
            1
            if self.last is not None and self.last.end_lineno is None
            else self.last.end_lineno + 1
        )
        self.last.extend(location, pcs=pcs, ws_start=start)

    def add_builtin_jump(
        self,
        name: str,
        _type: str,
        full_name: str | None = None,
        source_path: Path | None = None,
        pcs: set[int] | None = None,
    ):
        """
        A convenience method for appending a control flow that happened
        from an internal compiler built-in code. See the ape-vyper plugin
        for a usage example.

        Args:
            name (str): The name of the compiler built-in.
            _type (str): A str describing the type of check.
            full_name (str | None): A full-name ID.
            source_path (Path | None): The source file related, if there is one.
            pcs (set[int] | None): Program counter values mapping to this check.
        """
        pcs = pcs or set()
        closure = Closure(name=name, full_name=full_name or name)
        depth = self.last.depth - 1 if self.last else 0
        statement = Statement(type=_type, pcs=pcs)
        flow = ControlFlow(
            closure=closure, depth=depth, statements=[statement], source_path=source_path
        )
        self.append(flow)

    def _add(
        self,
        asts: list[ASTNode],
        content: Content,
        pcs: set[int],
        function: Function,
        depth: int,
        source_path: Path | None = None,
    ):
        statement = SourceStatement(asts=asts, content=content, pcs=pcs)
        exec_sequence = ControlFlow(
            statements=[statement], source_path=source_path, closure=function, depth=depth
        )
        self.append(exec_sequence)


@dataclass
class ContractFunctionPath:
    """
    Useful for identifying a method in a contract.
    """

    contract_name: str
    method_name: str | None = None

    @classmethod
    def from_str(cls, value: str) -> "ContractFunctionPath":
        if ":" in value:
            contract_name, method_name = value.split(":")
            return cls(contract_name=contract_name, method_name=method_name)

        return cls(contract_name=value)

    def __str__(self) -> str:
        return f"{self.contract_name}:{self.method_name}"

    @log_instead_of_fail(default="<ContractFunctionPath>")
    def __repr__(self) -> str:
        return f"<{self}>"



from collections.abc import Iterator
from typing import TYPE_CHECKING

from .pluggy_patch import PluginType, hookspec

if TYPE_CHECKING:
    from ape.api.convert import ConverterAPI


class ConversionPlugin(PluginType):
    """
    A plugin for converting values. The `ape-ens <https://github.com/ApeWorX/ape-ens>`__
    plugin is an example of a conversion-plugin.
    """

    @hookspec
    def converters(self) -> Iterator[tuple[str, type["ConverterAPI"]]]:  # type: ignore[empty-body]
        """
        A hook that returns an iterator of tuples of a string ABI type and a
        ``ConverterAPI`` subclass.

        Usage example::

            @plugins.register(plugins.ConversionPlugin)
            def converters():
                yield int, MweiConversions

        Returns:
            Iterator[tuple[str, type[:class:`~ape.api.convert.ConverterAPI`]]]
        """



from typing import NamedTuple

import pytest
from eth_pydantic_types import HexBytes
from ethpm_types import ContractType

from ape.exceptions import APINotImplementedError
from ape_ethereum.multicall import Call
from ape_ethereum.multicall.constants import MULTICALL3_ADDRESS, MULTICALL3_CONTRACT_TYPE
from ape_ethereum.multicall.exceptions import UnsupportedChainError

RETURNDATA = HexBytes("0x4a821464")


class ReturnData(NamedTuple):
    success: bool
    returnData: bytes


RETURNDATA_PARAMS = {
    "result_ok": (ReturnData(True, RETURNDATA), RETURNDATA),
    "result_fail": (ReturnData(False, RETURNDATA), None),
}


@pytest.fixture(scope="module")
def undeployed_multicall(chain):
    # NOTE: Still has the ability to decode/encode inputs.
    return chain.contracts.instance_at(
        MULTICALL3_ADDRESS,
        contract_type=ContractType.model_validate(MULTICALL3_CONTRACT_TYPE),
    )


@pytest.fixture(scope="module")
def aggregate3(undeployed_multicall):
    # NOTE: Avoid `__getattr__` call as it requires a real contract.
    return undeployed_multicall._mutable_methods_["aggregate3"]


def test_inject_raises_not_implemented():
    with pytest.raises(APINotImplementedError):
        Call.inject()


def test_unsupported_chain(call_handler_with_struct_input, struct_input_for_call):
    call = Call()
    call.add(call_handler_with_struct_input, *struct_input_for_call, allowFailure=True)

    # The local test chain doesn't have the multicall contracts.
    with pytest.raises(UnsupportedChainError):
        next(call())


def test_aggregate3_input(
    aggregate3,
    call_handler_with_struct_input,
    struct_input_for_call,
    vyper_contract_instance,
):
    call = Call()

    # Use a real contract here so the target encoding works.
    call_handler_with_struct_input.contract = vyper_contract_instance

    call.add(call_handler_with_struct_input, *struct_input_for_call)
    actual = aggregate3.encode_input(call.calls)
    assert isinstance(actual, HexBytes)


@pytest.mark.parametrize("returndata_key", RETURNDATA_PARAMS)
def test_returndata(returndata_key):
    result, output = RETURNDATA_PARAMS[returndata_key]
    call = Call()
    call._result = [result]  # type: ignore
    assert call.returnData[0] == output



from typing import Any

import pytest

from ape.api import ConverterAPI
from ape.exceptions import ConversionError


def test_convert_logs_and_passes_errors_from_is_convertible(conversion_manager, ape_caplog):
    """
    When checking if something is convertible, and is_convertible errors
    for whatever reason, log the error and consider it "not convertible".
    More than likely, it isn't by that converter and is a plugin-error.
    """
    error_msg = "Unexpected error!"

    class ProblematicConverter(ConverterAPI):
        def is_convertible(self, value: Any) -> bool:
            raise ValueError(error_msg)

        def convert(self, value: Any) -> Any:
            return value

    conversion_manager._converters[dict] = (ProblematicConverter(),)
    expected = f"Issue while checking `ProblematicConverter.is_convertible()`: {error_msg}"
    with pytest.raises(ConversionError):
        _ = conversion_manager.convert(123, dict)

    assert expected in ape_caplog.head



from typing import Any, Literal

from eth_typing import HexStr
from hexbytes import HexBytes

BlockID = int | HexStr | HexBytes | Literal["earliest", "latest", "pending"]
"""
An ID that can match a block, such as the literals ``"earliest"``, ``"latest"``, or ``"pending"``
as well as a block number or hash (HexBytes).
"""

ContractCode = str | bytes | HexBytes
"""
A type that represents contract code, which can be represented in string, bytes, or HexBytes.
"""

SnapshotID = Any
"""
An ID representing a point in time on a blockchain, as used in the
:meth:`~ape.managers.chain.ChainManager.snapshot` and
:meth:`~ape.managers.chain.ChainManager.snapshot` methods. Can be a ``str``, ``int``, ``bytes``,
``tuple[bytes, int]``, etc. Providers will expect and handle snapshot IDs differently.
There shouldn't be a need to change providers when using this feature, so there should not be
confusion over this type in practical use cases.
"""



import os
import shutil
import subprocess
import tempfile
import zipfile
from collections.abc import Iterator
from io import BytesIO
from pathlib import Path
from typing import Any

from requests import HTTPError, Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from ape.exceptions import CompilerError, ProjectError, UnknownVersionError
from ape.logging import logger
from ape.utils.misc import cached_property
from ape.utils.rpc import USER_AGENT, stream_response


class GitProcessWrapper:
    @cached_property
    def git(self) -> str:
        if path := shutil.which("git"):
            return path

        raise ProjectError("`git` not installed.")

    def clone(self, url: str, target_path: Path | None = None, branch: str | None = None):
        command = [self.git, "-c", "advice.detachedHead=false", "clone", url]

        if target_path:
            command.append(str(target_path))

        if branch is not None:
            command.extend(("--branch", branch))

        logger.debug(f"Running git command: '{' '.join(command)}'")
        result = subprocess.call(command)
        if result != 0:
            fail_msg = f"`git clone` command failed for '{url}'."

            if branch and not branch.startswith("v"):
                # Often times, `v` is required for tags.
                try:
                    self.clone(url, target_path, branch=f"v{branch}")
                except Exception:
                    raise ProjectError(fail_msg)

                # Succeeded when prefixing `v`.
                return

            # Failed and we don't really know why.
            # Shouldn't really happen.
            # User will have to run command separately to debug.
            raise ProjectError(fail_msg)


# NOTE: This client is only meant to be used internally for ApeWorX projects.
class _GithubClient:
    # Generic git/github client attributes.
    TOKEN_KEY = "GITHUB_ACCESS_TOKEN"
    API_URL_PREFIX = "https://api.github.com"
    git: GitProcessWrapper = GitProcessWrapper()

    # ApeWorX-specific attributes.
    ORGANIZATION_NAME = "ApeWorX"
    FRAMEWORK_NAME = "ape"
    _repo_cache: dict[str, dict] = {}

    def __init__(self, session: Session | None = None):
        if session:
            # NOTE: Mostly allowed for testing purposes.
            self.__session = session

        else:
            headers = {"Content-Type": "application/json", "User-Agent": USER_AGENT}
            if auth := os.environ[self.TOKEN_KEY] if self.TOKEN_KEY in os.environ else None:
                headers["Authorization"] = f"token {auth}"

            session = Session()
            session.headers = {**session.headers, **headers}
            adapter = HTTPAdapter(
                max_retries=Retry(total=10, backoff_factor=1.0, status_forcelist=[403]),
            )
            session.mount("https://", adapter)
            self.__session = session

    @cached_property
    def org(self) -> dict:
        """
        Our organization on ``Github``.
        """
        return self.get_organization(self.ORGANIZATION_NAME)

    @cached_property
    def available_plugins(self) -> set[str]:
        return {
            repo["name"].replace("-", "_")
            for repo in self.get_org_repos()
            if not repo.get("private", False) and repo["name"].startswith(f"{self.FRAMEWORK_NAME}-")
        }

    def get_org_repos(self) -> Iterator[dict]:
        params = {"per_page": 100, "page": 1}
        while True:
            response = self._get(f"orgs/{self.ORGANIZATION_NAME}/repos", params=params)
            repository_count = len(response)

            if repository_count == 0:
                break

            yield from response
            params["page"] += 1

    def get_release(self, org_name: str, repo_name: str, version: str) -> dict:
        if version == "latest":
            return self.get_latest_release(org_name, repo_name)

        def _try_get_release(vers):
            try:
                return self._get_release(org_name, repo_name, vers)
            except Exception:
                return None

        if release := _try_get_release(version):
            return release
        else:
            original_version = str(version)
            # Try an alternative tag style
            if version.startswith("v"):
                version = version.lstrip("v")
            else:
                version = f"v{version}"

            if release := _try_get_release(version):
                return release

            raise UnknownVersionError(original_version, repo_name)

    def _get_release(self, org_name: str, repo_name: str, version: str) -> dict:
        return self._get(f"repos/{org_name}/{repo_name}/releases/tags/{version}")

    def get_repo(self, org_name: str, repo_name: str) -> dict:
        repo_path = f"{org_name}/{repo_name}"
        if repo_path not in self._repo_cache:
            try:
                self._repo_cache[repo_path] = self._get_repo(org_name, repo_name)
                return self._repo_cache[repo_path]
            except Exception as err:
                raise ProjectError(f"Unknown repository '{repo_path}'") from err

        return self._repo_cache[repo_path]

    def _get_repo(self, org_name: str, repo_name: str) -> dict:
        try:
            return self._get(f"repos/{org_name}/{repo_name}")
        except HTTPError as err:
            if err.response.status_code == 404:
                raise ProjectError(f"Unknown repository '{org_name}/{repo_name}'")

            raise  # The original HTTPError

    def get_latest_release(self, org_name: str, repo_name: str) -> dict:
        return self._get(f"repos/{org_name}/{repo_name}/releases/latest")

    def get_organization(self, org_name: str) -> dict:
        return self._get(f"orgs/{org_name}")

    def clone_repo(
        self,
        org_name: str,
        repo_name: str,
        target_path: str | Path,
        branch: str | None = None,
        scheme: str = "https",
    ):
        repo = self.get_repo(org_name, repo_name)
        branch = branch or repo["default_branch"]
        logger.info(f"Cloning branch '{branch}' from '{repo['name']}'.")
        url = repo["git_url"]

        if "ssh" in scheme or "git" in scheme:
            url = url.replace("git://github.com/", "git@github.com:")
        elif "http" in scheme:
            url = url.replace("git://", "https://")
        else:
            raise ValueError(f"Scheme '{scheme}' not supported.")

        target_path = Path(target_path)
        target_path.parent.mkdir(parents=True, exist_ok=True)
        if target_path.exists():
            # Target repo cannot exist.
            target_path = target_path / repo_name

        self.git.clone(url, branch=branch, target_path=target_path)

    def download_package(
        self, org_name: str, repo_name: str, version: str, target_path: Path | str
    ):
        target_path = Path(target_path)  # Handles str
        if not target_path or not target_path.is_dir():
            raise ValueError(f"'target_path' must be a valid directory (got '{target_path}').")

        release = self.get_release(org_name, repo_name, version)
        description = f"Downloading {org_name}/{repo_name}@{version}"
        release_content = stream_response(
            release["zipball_url"], progress_bar_description=description
        )

        # Use temporary path to isolate a package when unzipping
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmp:
            temp_path = Path(tmp)
            with zipfile.ZipFile(BytesIO(release_content)) as zf:
                zf.extractall(temp_path)

            # Copy the directory contents into the target path.
            downloaded_packages = [f for f in temp_path.iterdir() if f.is_dir()]
            if len(downloaded_packages) < 1:
                raise CompilerError(f"Unable to download package at '{org_name}/{repo_name}'.")

            package_path = temp_path / downloaded_packages[0]
            for source_file in package_path.iterdir():
                shutil.move(str(source_file), str(target_path))

    def _get(self, url: str, params: dict | None = None) -> Any:
        return self._request("GET", url, params=params)

    def _request(self, method: str, url: str, **kwargs) -> Any:
        url = f"{self.API_URL_PREFIX}/{url}"
        response = self.__session.request(method, url, **kwargs)

        try:
            response.raise_for_status()
        except HTTPError as err:
            if err.response.status_code == 401 and self.__session.headers.get("Authorization"):
                token = self.__session.headers["Authorization"]
                del self.__session.headers["Authorization"]
                response = self.__session.request(method, url, **kwargs)
                try:
                    response.raise_for_status()  # Raise exception if the retry also fails
                except HTTPError:
                    # Even without the Authorization token, the request still failed.
                    # Raise the original error in this case. Also, put back token just in case.
                    self.__session.headers["Authorization"] = token
                    raise err
                else:
                    # The request failed with Authorization but succeeded without.
                    # Let the user know their token is likely expired.
                    logger.warning(
                        "Requests are not authorized! GITHUB_ACCESS_TOKEN is likely expired; "
                        "received 401 when attempted to use it. If you need GitHub authorization, "
                        "try resetting your token."
                    )
                    return response.json()

            raise  # Raise the error.

        return response.json()


github_client = _GithubClient()



import pytest

from ape.utils.misc import LOCAL_NETWORK_NAME
from tests.conftest import ApeSubprocessRunner, geth_process_test

from .utils import run_once, skip_projects_except

_DEFAULT_NETWORKS_TREE = """
ethereum  (default)
├── local  (default)
│   ├── node
│   └── test  (default)
├── mainnet
│   └── test  (default)
└── sepolia
    └── node  (default)
"""
_DEFAULT_NETWORKS_YAML = """
ecosystems:
- isDefault: true
  name: ethereum
  networks:
  - isDefault: true
    name: local
    providers:
    - name: node
    - isDefault: true
      name: test
  - name: mainnet
    providers:
    - isDefault: true
      name: node
  - name: mainnet-fork
    providers: []
  - name: sepolia
    providers:
    - isDefault: true
      name: node
  - name: sepolia-fork
    providers: []
"""
_NODE_NETWORKS_TREE = """
ethereum  (default)
├── local  (default)
│   └── node  (default)
├── mainnet
│   └── node  (default)
└── sepolia
    └── node  (default)
"""
_TEST_PROVIDER_TREE_OUTPUT = """
ethereum  (default)
└── local  (default)
    └── test  (default)
"""
_SEPOLIA_NETWORK_TREE_OUTPUT = """
ethereum  (default)
└── sepolia
    └── node  (default)
"""
_CUSTOM_NETWORKS_TREE = """
ethereum  (default)
├── apenet
│   └── node  (default)
├── apenet1
│   └── node  (default)
├── local  (default)
│   └── node  (default)
├── mainnet
│   └── node  (default)
└── sepolia
    └── node  (default)
"""


@pytest.fixture
def networks_runner(config):
    class NetworksSubprocessRunner(ApeSubprocessRunner):
        def __init__(self):
            super().__init__("networks", data_folder=config.DATA_FOLDER)

    return NetworksSubprocessRunner()


def assert_rich_text(actual: str, expected: str):
    """
    The output from `rich` causes a bunch of extra spaces to
    appear at the end of each line. For easier testing, we remove those here.
    Also, we ignore whether the expected line is at the end or in the middle
    of the output to handle cases when the test-runner has additional plugins
    installed.
    """
    expected_lines = [
        x.replace("└", "").replace("├", "").replace("│", "").strip()
        for x in expected.strip().split("\n")
    ]
    actual_lines = [
        x.replace("└", "").replace("├", "").replace("│", "").strip()
        for x in actual.strip().split("\n")
    ]

    for expected_line in expected_lines:
        assert expected_line in actual_lines


@run_once
def test_list(ape_cli, runner):
    result = runner.invoke(ape_cli, ("networks", "list"))
    assert result.exit_code == 0

    # Grab ethereum
    actual = "ethereum  (default)\n" + "".join(result.output.split("ethereum  (default)\n")[-1])

    assert_rich_text(actual, _DEFAULT_NETWORKS_TREE)


@run_once
def test_list_yaml(ape_cli, runner):
    result = runner.invoke(ape_cli, ("networks", "list", "--format", "yaml"))
    expected_lines = _DEFAULT_NETWORKS_YAML.strip().split("\n")

    for expected_line in expected_lines:
        if expected_line.lstrip() == "providers: []":
            # Skip these lines in case test-runner has installed providers
            continue

        if (
            expected_line.lstrip().startswith("- name:")
            and expected_line not in result.output
            and "explorer:" in result.output
        ):
            # May have explorers installed - ignore that.
            expected_line = expected_line.lstrip(" -")

        assert expected_line in result.output, result.output


@skip_projects_except("geth")
def test_list_geth(ape_cli, runner, networks, project):
    result = runner.invoke(ape_cli, ("networks", "list"))
    assert result.exit_code == 0

    # Grab ethereum
    actual = "ethereum  (default)\n" + "".join(result.output.split("ethereum  (default)\n")[-1])

    assert_rich_text(actual, _NODE_NETWORKS_TREE)

    # Assert that URI still exists for local network
    # (was bug where one network's URI disappeared when setting different network's URI)
    geth_provider = networks.get_provider_from_choice(f"ethereum:{LOCAL_NETWORK_NAME}:node")
    actual_uri = geth_provider.uri
    assert actual_uri.startswith("http")


@geth_process_test
@skip_projects_except("geth")
def test_list_running(ape_cli, runner, geth_provider):
    result = runner.invoke(ape_cli, ("networks", "list", "--running"))
    assert result.exit_code == 0
    assert geth_provider.ipc_path is not None, "any uri is needed for test"
    actual = "".join(result.output.split("\n"))
    assert f"{geth_provider.ipc_path}" in actual or "Local node(s) not running." in actual


@run_once
def test_list_filter_networks(ape_cli, runner):
    result = runner.invoke(ape_cli, ("networks", "list", "--network", "sepolia"))
    assert result.exit_code == 0

    # Grab ethereum
    actual = "ethereum  (default)\n" + "".join(result.output.split("ethereum  (default)\n")[-1])

    assert_rich_text(actual, _SEPOLIA_NETWORK_TREE_OUTPUT)


@run_once
def test_list_filter_providers(ape_cli, runner, networks):
    result = runner.invoke(ape_cli, ("networks", "list", "--provider", "test"))
    assert result.exit_code == 0

    # Grab ethereum
    actual = "ethereum  (default)\n" + "".join(result.output.split("ethereum  (default)\n")[-1])

    assert_rich_text(actual, _TEST_PROVIDER_TREE_OUTPUT)


@skip_projects_except("geth")
def test_list_custom_networks(integ_project, networks_runner):
    networks_runner.project = integ_project
    result = networks_runner.invoke("list")
    assert result.exit_code == 0
    actual = "ethereum  (default)\n" + "".join(result.output.split("ethereum  (default)\n")[-1])
    assert_rich_text(actual, _CUSTOM_NETWORKS_TREE)


@run_once
def test_run_not_subprocess_provider(networks_runner):
    cmd = ("run", "--network", "ethereum:local:test")
    result = networks_runner.invoke(*cmd)
    expected = "`ape networks run` requires a provider that manages a process, not 'test'."
    assert result.exit_code != 0
    assert expected in result.output


@run_once
def test_run_custom_network(ape_cli, runner):
    cmd = ("networks", "run", "--network", "ethereum:local:test")
    result = runner.invoke(ape_cli, cmd)
    expected = "`ape networks run` requires a provider that manages a process, not 'test'"
    assert result.exit_code != 0
    assert expected in result.output


@run_once
def test_run_block_time(ape_cli, runner):
    cmd = (
        "networks",
        "run",
        "--network",
        "ethereum:local:test",
        "--block-time",
        "10",
    )
    result = runner.invoke(ape_cli, cmd)
    expected = "`ape networks run` requires a provider that manages a process, not 'test'"
    assert result.exit_code != 0
    assert expected in result.output


@geth_process_test
@skip_projects_except("geth")
def test_run_already_running(networks_runner, integ_project, geth_provider):
    networks_runner.project = integ_project
    cmd = ("run", "--network", f"ethereum:{LOCAL_NETWORK_NAME}:node")
    result = networks_runner.invoke(*cmd)
    assert result.exit_code != 0
    assert "ERROR" in result.output
    assert "Process already running." in result.output



import re
from decimal import Decimal

from ape.api.convert import ConverterAPI
from ape.types.units import CurrencyValue

ETHER_UNITS = {
    "eth": int(1e18),
    "ether": int(1e18),
    "milliether": int(1e15),
    "finney": int(1e15),
    "microether": int(1e12),
    "szabo": int(1e12),
    "gwei": int(1e9),
    "shannon": int(1e9),
    "mwei": int(1e6),
    "lovelace": int(1e6),
    "kwei": int(1e3),
    "babbage": int(1e3),
    "wei": 1,
}
NUMBER_PATTERN = re.compile(r"^-?\d{1,3}(?:[,_]?\d{3})*(?:\.\d+)?(?:[eE][+-]?\d+)?$")


class WeiConversions(ConverterAPI):
    """Converts units like `1 ether` to 1e18 wei."""

    def is_convertible(self, value: str) -> bool:
        if not isinstance(value, str):
            return False

        elif " " not in value or len(value.split(" ")) > 2:
            return False

        val, unit = value.split(" ")
        return unit.lower() in ETHER_UNITS and bool(NUMBER_PATTERN.match(val))

    def convert(self, value: str) -> int:
        value, unit = value.split(" ")
        converted_value = int(
            Decimal(value.replace("_", "").replace(",", "")) * ETHER_UNITS[unit.lower()]
        )
        return CurrencyValue(converted_value)



from tests.conftest import geth_process_test


@geth_process_test
def test_get_contract_metadata(
    mock_geth, geth_contract, geth_account, chain, networks, geth_provider
):
    networks.active_provider = mock_geth
    actual = chain.contracts.get_creation_metadata(geth_contract.address)
    assert actual.deployer == geth_account.address

    # hold onto block, setup mock.
    block = geth_provider.get_block(actual.block)
    del chain.contracts.contract_creations[geth_contract.address]
    mock_geth.web3.eth.get_block.return_value = block

    orig_web3 = chain.network_manager.active_provider._web3
    chain.network_manager.active_provider._web3 = mock_geth.web3
    try:
        for client in ("geth", "erigon"):
            chain.network_manager.active_provider._client_version = client
            _ = chain.contracts.get_creation_metadata(geth_contract.address)
    finally:
        chain.network_manager.active_provider._web3 = orig_web3

    call_args = mock_geth._web3.provider.make_request.call_args_list

    # geth
    assert call_args[-2][0][0] == "debug_traceBlockByNumber"
    assert call_args[-2][0][1][1] == {"tracer": "callTracer"}
    # parity
    assert call_args[-1][0][0] == "trace_replayBlockTransactions"
    assert call_args[-1][0][1][1] == ["trace"]
